"""jc - JSON Convert `/proc/<pid>/maps` file parser

Usage (cli):

    $ cat /proc/1/maps | jc --proc

or

    $ jc /proc/1/maps

or

    $ cat /proc/1/maps | jc --proc-pid-maps

Usage (module):

    import jc
    result = jc.parse('proc', proc_pid_maps_file)

or

    import jc
    result = jc.parse('proc_pid_maps', proc_pid_maps_file)

Schema:

    [
      {
        "start":                          string,
        "end":                            string,
        "perms": [
                                          string
        ],
        "offset":                         string,
        "inode":                          integer,
        "pathname":                       string,
        "maj":                            string,
        "min":                            string
      }
    ]

Examples:

    $ cat /proc/1/maps | jc --proc -p
    [
      {
        "perms": [
          "read",
          "private"
        ],
        "offset": "00000000",
        "inode": 798126,
        "pathname": "/usr/lib/systemd/systemd",
        "start": "55a9e753c000",
        "end": "55a9e7570000",
        "maj": "fd",
        "min": "00"
      },
      {
        "perms": [
          "read",
          "execute",
          "private"
        ],
        "offset": "00034000",
        "inode": 798126,
        "pathname": "/usr/lib/systemd/systemd",
        "start": "55a9e7570000",
        "end": "55a9e763a000",
        "maj": "fd",
        "min": "00"
      },
      ...
    ]

    $ cat /proc/1/maps | jc --proc-pid-maps -p -r
    [
      {
        "address": "55a9e753c000-55a9e7570000",
        "perms": "r--p",
        "offset": "00000000",
        "dev": "fd:00",
        "inode": "798126",
        "pathname": "/usr/lib/systemd/systemd"
      },
      {
        "address": "55a9e7570000-55a9e763a000",
        "perms": "r-xp",
        "offset": "00034000",
        "dev": "fd:00",
        "inode": "798126",
        "pathname": "/usr/lib/systemd/systemd"
      },
      ...
    ]
"""
from typing import List, Dict
import jc.utils
from jc.parsers.universal import simple_table_parse


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/<pid>/maps` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'inode'}

    perms_map = {
        'r': 'read',
        'w': 'write',
        'x': 'execute',
        's': 'shared',
        'p': 'private',
        '-': None
    }

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = int(entry[key])

        if 'address' in entry:
            start, end = entry['address'].split('-')
            entry['start'] = start
            entry['end'] = end
            del entry['address']

        if 'perms' in entry:
            perms_list = [perms_map[x] for x in entry['perms'] if perms_map[x]]
            entry['perms'] = perms_list

        if 'dev' in entry:
            maj, min = entry['dev'].split(':', maxsplit=1)
            entry['maj'] = maj
            entry['min'] = min
            del entry['dev']

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []

    if jc.utils.has_data(data):

        header = 'address           perms offset  dev   inode      pathname\n'
        data = header + data
        raw_output = simple_table_parse(data.splitlines())

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert `env` and `printenv` command output parser

This parser will output a list of dictionaries each containing `name` and
`value` keys. If you would like a simple dictionary output, then use the
`-r` command-line option or the `raw=True` argument in the `parse()`
function.

Usage (cli):

    $ env | jc --env

or

    $ jc env

Usage (module):

    import jc
    result = jc.parse('env', env_command_output)

Schema:

    [
      {
        "name":     string,
        "value":    string
      }
    ]

Examples:

    $ env | jc --env -p
    [
      {
        "name": "XDG_SESSION_ID",
        "value": "1"
      },
      {
        "name": "HOSTNAME",
        "value": "localhost.localdomain"
      },
      {
        "name": "TERM",
        "value": "vt220"
      },
      {
        "name": "SHELL",
        "value": "/bin/bash"
      },
      {
        "name": "HISTSIZE",
        "value": "1000"
      },
      ...
    ]

    $ env | jc --env -p -r
    {
      "TERM": "xterm-256color",
      "SHELL": "/bin/bash",
      "USER": "root",
      "PATH": "/usr/local/bin:/usr/bin:/bin:/usr/sbin:/sbin",
      "PWD": "/root",
      "LANG": "en_US.UTF-8",
      "HOME": "/root",
      "LOGNAME": "root",
      "_": "/usr/bin/env"
    }
"""
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.4'
    description = '`env` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    magic_commands = ['env', 'printenv']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """

    # rebuild output for added semantic information
    processed = []
    for k, v in proc_data.items():
        proc_line = {}
        proc_line['name'] = k
        proc_line['value'] = v
        processed.append(proc_line)

    return processed


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary of raw structured data or
        List of Dictionaries of processed structured data
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = {}

    # Clear any blank lines
    cleandata = list(filter(None, data.splitlines()))

    if jc.utils.has_data(data):

        for entry in cleandata:
            parsed_line = entry.split('=', maxsplit=1)
            raw_output[parsed_line[0]] = parsed_line[1]

    if raw:
        return raw_output
    else:
        return _process(raw_output)



"""jc - JSON Convert `update-alternatives --query` command output parser

Usage (cli):

    $ update-alternatives --query | jc --update-alt-q

or

    $ jc update-alternatives --query

Usage (module):

    import jc
    result = jc.parse('update_alt_q',
                      update_alternatives_query_command_output)

Schema:

    {
      "name":                 string,
      "link":                 string,
      "slaves": [
        {
          "name":             string,
          "path":             string
        }
      ],
      "status":               string,
      "best":                 string,
      "value":                string,   # (null if 'none')
      "alternatives": [
        {
          "alternative":      string,
          "priority":         integer,
          "slaves": [
            {
              "name":         string,
              "path":         string
            }
          ]
        }
      ]
    }

Examples:

    $ update-alternatives --query editor | jc --update-alt-q -p
    {
      "name": "editor",
      "link": "/usr/bin/editor",
      "slaves": [
        {
          "name": "editor.1.gz",
          "path": "/usr/share/man/man1/editor.1.gz"
        },
        {
          "name": "editor.da.1.gz",
          "path": "/usr/share/man/da/man1/editor.1.gz"
        }
      ],
      "status": "auto",
      "best": "/bin/nano",
      "value": "/bin/nano",
      "alternatives": [
        {
          "alternative": "/bin/ed",
          "priority": -100,
          "slaves": [
            {
              "name": "editor.1.gz",
              "path": "/usr/share/man/man1/ed.1.gz"
            }
          ]
        },
        {
          "alternative": "/bin/nano",
          "priority": 40,
          "slaves": [
            {
              "name": "editor.1.gz",
              "path": "/usr/share/man/man1/nano.1.gz"
            }
          ]
        }
      ]
    }

    $ update-alternatives --query | jc --update-alt-q -p -r
    {
      "name": "editor",
      "link": "/usr/bin/editor",
      "slaves": [
        {
          "name": "editor.1.gz",
          "path": "/usr/share/man/man1/editor.1.gz"
        },
        {
          "name": "editor.da.1.gz",
          "path": "/usr/share/man/da/man1/editor.1.gz"
        }
      ],
      "status": "auto",
      "best": "/bin/nano",
      "value": "/bin/nano",
      "alternatives": [
        {
          "alternative": "/bin/ed",
          "priority": "-100",
          "slaves": [
            {
              "name": "editor.1.gz",
              "path": "/usr/share/man/man1/ed.1.gz"
            }
          ]
        },
        {
          "alternative": "/bin/nano",
          "priority": "40",
          "slaves": [
            {
              "name": "editor.1.gz",
              "path": "/usr/share/man/man1/nano.1.gz"
            }
          ]
        }
      ]
    }
"""
from typing import List, Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.2'
    description = '`update-alternatives --query` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    magic_commands = ['update-alternatives --query']
    tags = ['command']


__version__ = info.version


def _process(proc_data: Dict) -> Dict:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary. Structured to conform to the schema.
    """
    int_list = {'priority'}

    if 'value' in proc_data:
        if proc_data['value'] == 'none':
            proc_data['value'] = None

    if 'alternatives' in proc_data:
        for index, alt in enumerate(proc_data['alternatives']):
            for key in alt:
                if key in int_list:
                    proc_data['alternatives'][index][key] = jc.utils.convert_to_int(proc_data['alternatives'][index][key])

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> Dict:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: Dict = {}
    slaves: List = []
    alt_obj: Dict = {}

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):
            line_list = line.split(maxsplit=1)

            if line.startswith('Name: '):
                raw_output['name'] = line_list[1]
                continue

            if line.startswith('Link: '):
                raw_output['link'] = line_list[1]
                continue

            if line.startswith('Slaves:'):
                continue

            if line.startswith(' '):
                s_name = line_list[0].strip()
                s_path = line_list[1]
                slaves.append(
                    {
                        "name": s_name,
                        "path": s_path
                    }
                )
                continue

            if line.startswith('Status: '):
                if slaves:
                    raw_output['slaves'] = slaves
                    slaves = []
                raw_output['status'] = line_list[1]
                continue

            if line.startswith('Best: '):
                raw_output['best'] = line_list[1]
                continue

            if line.startswith('Value: '):
                raw_output['value'] = line_list[1]
                continue

            if line.startswith('Alternative: '):
                if not 'alternatives' in raw_output:
                    raw_output['alternatives'] = []

                if alt_obj:
                    if slaves:
                        alt_obj['slaves'] = slaves
                        slaves = []

                    raw_output['alternatives'].append(alt_obj)

                alt_obj = {"alternative": line_list[1]}
                continue

            if line.startswith('Priority: '):
                alt_obj['priority'] = line_list[1]
                continue

        if alt_obj:
            if slaves:
                alt_obj['slaves'] = slaves
            raw_output['alternatives'].append(alt_obj)

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert `ssh` configuration file and `ssh -G` command output parser

This parser will work with `ssh` configuration files or the output of
`ssh -G`. Any `Match` blocks in the `ssh` configuration file will be
ignored.

Usage (cli):

    $ ssh -G hostname | jc --ssh-conf

or

    $ jc ssh -G hostname

or

    $ cat ~/.ssh/config | jc --ssh-conf

Usage (module):

    import jc
    result = jc.parse('ssh_conf', ssh_conf_output)

Schema:

    [
      {
        "host":                                       string,
        "host_list": [
                                                      string
        ],
        "addkeystoagent":                             string,
        "addressfamily":                              string,
        "batchmode":                                  string,
        "bindaddress":                                string,
        "bindinterface":                              string,
        "canonicaldomains": [
                                                      string
        ],
        "canonicalizefallbacklocal":                  string,
        "canonicalizehostname":                       string,
        "canonicalizemaxdots":                        integer,
        "canonicalizepermittedcnames": [
                                                      string
        ],
        "casignaturealgorithms": [
                                                      string
        ],
        "certificatefile": [
                                                      string
        ],
        "checkhostip":                                string,
        "ciphers": [
                                                      string
        ],
        "clearallforwardings":                        string,
        "compression":                                string,
        "connectionattempts":                         integer,
        "connecttimeout":                             integer,
        "controlmaster":                              string,
        "controlpath":                                string,
        "controlpersist":                             string,
        "dynamicforward":                             string,
        "enableescapecommandline":                    string,
        "enablesshkeysign":                           string,
        "escapechar":                                 string,
        "exitonforwardfailure":                       string,
        "fingerprinthash":                            string,
        "forkafterauthentication":                    string,
        "forwardagent":                               string,
        "forwardx11":                                 string,
        "forwardx11timeout":                          integer,
        "forwardx11trusted":                          string,
        "gatewayports":                               string,
        "globalknownhostsfile": [
                                                      string
        ],
        "gssapiauthentication":                       string,
        "gssapidelegatecredentials":                  string,
        "hashknownhosts":                             string,
        "hostbasedacceptedalgorithms": [
                                                      string
        ],
        "hostbasedauthentication":                    string,
        "hostkeyalgorithms": [
                                                      string
        ],
        "hostkeyalias":                               string,
        "hostname":                                   string,
        "identitiesonly":                             string,
        "identityagent":                              string,
        "identityfile": [
                                                      string
        ],
        "ignoreunknown":                              string,
        "include": [
                                                      string
        ],
        "ipqos": [
                                                      string
        ],
        "kbdinteractiveauthentication":               string,
        "kbdinteractivedevices": [
                                                      string
        ],
        "kexalgorithms": [
                                                      string
        ],
        "kexalgorithms_strategy":                     string,
        "knownhostscommand":                          string,
        "localcommand":                               string,
        "localforward": [
                                                      string
        ],
        "loglevel":                                   string,
        "logverbose": [
                                                      string
        ],
        "macs": [
                                                      string
        ],
        "macs_strategy":                              string,
        "nohostauthenticationforlocalhost":           string,
        "numberofpasswordprompts":                    integer,
        "passwordauthentication":                     string,
        "permitlocalcommand":                         string,
        "permitremoteopen": [
                                                      string
        ],
        "pkcs11provider":                             string,
        "port":                                       integer,
        "preferredauthentications": [
                                                      string
        ],
        "protocol":                                   integer,
        "proxycommand":                               string,
        "proxyjump": [
                                                      string
        ],
        "proxyusefdpass":                             string,
        "pubkeyacceptedalgorithms": [
                                                      string
        ],
        "pubkeyacceptedalgorithms_strategy":          string,
        "pubkeyauthentication":                       string,
        "rekeylimit":                                 string,
        "remotecommand":                              string,
        "remoteforward":                              string,
        "requesttty":                                 string,
        "requiredrsasize":                            integer,
        "revokedhostkeys":                            string,
        "securitykeyprovider":                        string,
        "sendenv": [
                                                      string
        ],
        "serveralivecountmax":                        integer,
        "serveraliveinterval":                        integer,
        "sessiontype":                                string,
        "setenv": [
                                                      string
        ],
        "stdinnull":                                  string,
        "streamlocalbindmask":                        string,
        "streamlocalbindunlink":                      string,
        "stricthostkeychecking":                      string,
        "syslogfacility":                             string,
        "tcpkeepalive":                               string,
        "tunnel":                                     string,
        "tunneldevice":                               string,
        "updatehostkeys":                             string,
        "user":                                       string,
        "userknownhostsfile": [
                                                      string
        ],
        "verifyhostkeydns":                           string,
        "visualhostkey":                              string,
        "xauthlocation":                              string
      }
    ]

Examples:

    $ ssh -G - | jc --ssh-conf -p
    [
      {
        "user": "foo",
        "hostname": "-",
        "port": 22,
        "addressfamily": "any",
        "batchmode": "no",
        "canonicalizefallbacklocal": "yes",
        "canonicalizehostname": "false",
        "checkhostip": "no",
        "compression": "no",
        "controlmaster": "false",
        "enablesshkeysign": "no",
        "clearallforwardings": "no",
        "exitonforwardfailure": "no",
        "fingerprinthash": "SHA256",
        "forwardx11": "no",
        "forwardx11trusted": "no",
        "gatewayports": "no",
        "gssapiauthentication": "no",
        "gssapidelegatecredentials": "no",
        "hashknownhosts": "no",
        "hostbasedauthentication": "no",
        "identitiesonly": "no",
        "kbdinteractiveauthentication": "yes",
        "nohostauthenticationforlocalhost": "no",
        "passwordauthentication": "yes",
        "permitlocalcommand": "no",
        "proxyusefdpass": "no",
        "pubkeyauthentication": "true",
        "requesttty": "auto",
        "sessiontype": "default",
        "stdinnull": "no",
        "forkafterauthentication": "no",
        "streamlocalbindunlink": "no",
        "stricthostkeychecking": "ask",
        "tcpkeepalive": "yes",
        "tunnel": "false",
        "verifyhostkeydns": "false",
        "visualhostkey": "no",
        "updatehostkeys": "true",
        "applemultipath": "no",
        "canonicalizemaxdots": 1,
        "connectionattempts": 1,
        "forwardx11timeout": 1200,
        "numberofpasswordprompts": 3,
        "serveralivecountmax": 3,
        "serveraliveinterval": 0,
        "ciphers": [
          "chacha20-poly1305@openssh.com",
          "aes128-ctr",
          "aes192-ctr",
          "aes256-ctr",
          "aes128-gcm@openssh.com",
          "aes256-gcm@openssh.com"
        ],
        "hostkeyalgorithms": [
          "ssh-ed25519-cert-v01@openssh.com",
          "ecdsa-sha2-nistp256-cert-v01@openssh.com",
          "ecdsa-sha2-nistp384-cert-v01@openssh.com",
          "ecdsa-sha2-nistp521-cert-v01@openssh.com",
          "rsa-sha2-512-cert-v01@openssh.com",
          "rsa-sha2-256-cert-v01@openssh.com",
          "ssh-ed25519",
          "ecdsa-sha2-nistp256",
          "ecdsa-sha2-nistp384",
          "ecdsa-sha2-nistp521",
          "rsa-sha2-512",
          "rsa-sha2-256"
        ],
        "hostbasedacceptedalgorithms": [
          "ssh-ed25519-cert-v01@openssh.com",
          "ecdsa-sha2-nistp256-cert-v01@openssh.com",
          "ecdsa-sha2-nistp384-cert-v01@openssh.com",
          "ecdsa-sha2-nistp521-cert-v01@openssh.com",
          "rsa-sha2-512-cert-v01@openssh.com",
          "rsa-sha2-256-cert-v01@openssh.com",
          "ssh-ed25519",
          "ecdsa-sha2-nistp256",
          "ecdsa-sha2-nistp384",
          "ecdsa-sha2-nistp521",
          "rsa-sha2-512",
          "rsa-sha2-256"
        ],
        "kexalgorithms": [
          "sntrup761x25519-sha512@openssh.com",
          "curve25519-sha256",
          "curve25519-sha256@libssh.org",
          "ecdh-sha2-nistp256",
          "ecdh-sha2-nistp384",
          "ecdh-sha2-nistp521",
          "diffie-hellman-group-exchange-sha256",
          "diffie-hellman-group16-sha512",
          "diffie-hellman-group18-sha512",
          "diffie-hellman-group14-sha256"
        ],
        "casignaturealgorithms": [
          "ssh-ed25519",
          "ecdsa-sha2-nistp256",
          "ecdsa-sha2-nistp384",
          "ecdsa-sha2-nistp521",
          "rsa-sha2-512",
          "rsa-sha2-256"
        ],
        "loglevel": "INFO",
        "macs": [
          "umac-64-etm@openssh.com",
          "umac-128-etm@openssh.com",
          "hmac-sha2-256-etm@openssh.com",
          "hmac-sha2-512-etm@openssh.com",
          "hmac-sha1-etm@openssh.com",
          "umac-64@openssh.com",
          "umac-128@openssh.com",
          "hmac-sha2-256",
          "hmac-sha2-512",
          "hmac-sha1"
        ],
        "securitykeyprovider": "$SSH_SK_PROVIDER",
        "pubkeyacceptedalgorithms": [
          "ssh-ed25519-cert-v01@openssh.com",
          "ecdsa-sha2-nistp256-cert-v01@openssh.com",
          "ecdsa-sha2-nistp384-cert-v01@openssh.com",
          "ecdsa-sha2-nistp521-cert-v01@openssh.com",
          "rsa-sha2-512-cert-v01@openssh.com",
          "rsa-sha2-256-cert-v01@openssh.com",
          "ssh-ed25519",
          "ecdsa-sha2-nistp256",
          "ecdsa-sha2-nistp384",
          "ecdsa-sha2-nistp521",
          "rsa-sha2-512",
          "rsa-sha2-256"
        ],
        "xauthlocation": "/usr/X11R6/bin/xauth",
        "identityfile": [
          "~/.ssh/id_rsa",
          "~/.ssh/id_ecdsa",
          "~/.ssh/id_ecdsa_sk",
          "~/.ssh/id_ed25519",
          "~/.ssh/id_ed25519_sk",
          "~/.ssh/id_xmss",
          "~/.ssh/id_dsa"
        ],
        "canonicaldomains": [
          "none"
        ],
        "globalknownhostsfile": [
          "/etc/ssh/ssh_known_hosts",
          "/etc/ssh/ssh_known_hosts2"
        ],
        "userknownhostsfile": [
          "/Users/foo/.ssh/known_hosts",
          "/Users/foo/.ssh/known_hosts2"
        ],
        "sendenv": [
          "LANG",
          "LC_*"
        ],
        "logverbose": [
          "none"
        ],
        "permitremoteopen": [
          "any"
        ],
        "addkeystoagent": "false",
        "forwardagent": "no",
        "connecttimeout": null,
        "tunneldevice": "any:any",
        "canonicalizepermittedcnames": [
          "none"
        ],
        "controlpersist": "no",
        "escapechar": "~",
        "ipqos": [
          "af21",
          "cs1"
        ],
        "rekeylimit": "0 0",
        "streamlocalbindmask": "0177",
        "syslogfacility": "USER"
      }
    ]

    $ cat ~/.ssh/config | jc --ssh-conf -p
    [
      {
        "host": "server1",
        "host_list": [
          "server1"
        ],
        "hostname": "server1.cyberciti.biz",
        "user": "nixcraft",
        "port": 4242,
        "identityfile": [
          "/nfs/shared/users/nixcraft/keys/server1/id_rsa"
        ]
      },
      {
        "host": "nas01",
        "host_list": [
          "nas01"
        ],
        "hostname": "192.168.1.100",
        "user": "root",
        "identityfile": [
          "~/.ssh/nas01.key"
        ]
      },
      {
        "host": "aws.apache",
        "host_list": [
          "aws.apache"
        ],
        "hostname": "1.2.3.4",
        "user": "wwwdata",
        "identityfile": [
          "~/.ssh/aws.apache.key"
        ]
      },
      {
        "host": "uk.gw.lan uk.lan",
        "host_list": [
          "uk.gw.lan",
          "uk.lan"
        ],
        "hostname": "192.168.0.251",
        "user": "nixcraft",
        "proxycommand": "ssh nixcraft@gateway.uk.cyberciti.biz nc %h %p 2> /dev/null"
      },
      {
        "host": "proxyus",
        "host_list": [
          "proxyus"
        ],
        "hostname": "vps1.cyberciti.biz",
        "user": "breakfree",
        "identityfile": [
          "~/.ssh/vps1.cyberciti.biz.key"
        ],
        "localforward": [
          "3128 127.0.0.1:3128"
        ]
      },
      {
        "host": "*",
        "host_list": [
          "*"
        ],
        "forwardagent": "no",
        "forwardx11": "no",
        "forwardx11trusted": "yes",
        "user": "nixcraft",
        "port": 22,
        "protocol": 2,
        "serveraliveinterval": 60,
        "serveralivecountmax": 30
      }
    ]

    $ cat ~/.ssh/config | jc --ssh-conf -p -r
    [
      {
        "host": "server1",
        "host_list": [
          "server1"
        ],
        "hostname": "server1.cyberciti.biz",
        "user": "nixcraft",
        "port": "4242",
        "identityfile": [
          "/nfs/shared/users/nixcraft/keys/server1/id_rsa"
        ]
      },
      {
        "host": "nas01",
        "host_list": [
          "nas01"
        ],
        "hostname": "192.168.1.100",
        "user": "root",
        "identityfile": [
          "~/.ssh/nas01.key"
        ]
      },
      {
        "host": "aws.apache",
        "host_list": [
          "aws.apache"
        ],
        "hostname": "1.2.3.4",
        "user": "wwwdata",
        "identityfile": [
          "~/.ssh/aws.apache.key"
        ]
      },
      {
        "host": "uk.gw.lan uk.lan",
        "host_list": [
          "uk.gw.lan",
          "uk.lan"
        ],
        "hostname": "192.168.0.251",
        "user": "nixcraft",
        "proxycommand": "ssh nixcraft@gateway.uk.cyberciti.biz nc %h %p 2> /dev/null"
      },
      {
        "host": "proxyus",
        "host_list": [
          "proxyus"
        ],
        "hostname": "vps1.cyberciti.biz",
        "user": "breakfree",
        "identityfile": [
          "~/.ssh/vps1.cyberciti.biz.key"
        ],
        "localforward": [
          "3128 127.0.0.1:3128"
        ]
      },
      {
        "host": "*",
        "host_list": [
          "*"
        ],
        "forwardagent": "no",
        "forwardx11": "no",
        "forwardx11trusted": "yes",
        "user": "nixcraft",
        "port": "22",
        "protocol": "2",
        "serveraliveinterval": "60",
        "serveralivecountmax": "30"
      }
    ]
"""
from typing import Set, List, Dict
from jc.jc_types import JSONDictType
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`ssh` config file and `ssh -G` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'freebsd']
    magic_commands = ['ssh -G']
    tags = ['command', 'file']


__version__ = info.version


def _process(proc_data: List[JSONDictType]) -> List[JSONDictType]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    split_fields_space: Set[str] = {
        'canonicaldomains', 'globalknownhostsfile', 'include', 'ipqos',
        'permitremoteopen', 'sendenv', 'setenv', 'userknownhostsfile'
    }

    split_fields_comma: Set[str] = {
        'canonicalizepermittedcnames', 'casignaturealgorithms', 'ciphers',
        'hostbasedacceptedalgorithms', 'hostkeyalgorithms',
        'kbdinteractivedevices', 'kexalgorithms', 'logverbose', 'macs',
        'preferredauthentications', 'proxyjump', 'pubkeyacceptedalgorithms'
    }

    int_list: Set[str] = {
        'canonicalizemaxdots', 'connectionattempts', 'connecttimeout',
        'forwardx11timeout', 'numberofpasswordprompts', 'port', 'protocol',
        'requiredrsasize', 'serveralivecountmax', 'serveraliveinterval'
    }

    for host in proc_data:
        dict_copy = host.copy()
        for key, val in dict_copy.items():
            # these are list values
            if key == 'sendenv' or key == 'setenv' or key == 'include':
                new_list: List[str] = []
                for item in val:
                    new_list.extend(item.split())
                host[key] = new_list
                continue

            if key in split_fields_space:
                host[key] = val.split()
                continue

            if key in split_fields_comma:
                host[key] = val.split(',')
                continue

        for key, val in host.items():
            if key in int_list:
                host[key] = jc.utils.convert_to_int(val)

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []
    host: Dict = {}

    multi_fields: Set[str] = {
        'certificatefile', 'identityfile', 'include', 'localforward',
        'sendenv', 'setenv'
    }

    modified_fields: Set[str] = {
        'casignaturealgorithms', 'ciphers', 'hostbasedacceptedalgorithms',
        'HostKeyAlgorithms', 'kexalgorithms', 'macs',
        'pubkeyacceptedalgorithms'
    }

    modifiers: Set[str] = {'+', '-', '^'}

    match_block_found = False

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):
            # skip any lines with only whitespace
            if not line.strip():
                continue

            # support configuration file by skipping commented lines
            if line.strip().startswith('#'):
                continue

            if line.strip().startswith('Host '):
                if host:
                    raw_output.append(host)

                hostnames = line.split(maxsplit=1)[1]
                host = {
                    'host': hostnames,
                    'host_list': hostnames.split()
                }

            # support configuration file by ignoring all lines between
            # Match xxx and Match any
            if line.strip().startswith('Match all'):
                match_block_found = False
                continue

            if line.strip().startswith('Match'):
                match_block_found = True
                continue

            if match_block_found:
                continue

            key, val = line.split(maxsplit=1)

            # support configuration file by converting to lower case
            key = key.lower()

            if key in multi_fields:
                if key not in host:
                    host[key] = []
                host[key].append(val)
                continue

            if key in modified_fields and val[0] in modifiers:
                host[key] = val[1:]
                host[key + '_strategy'] = val[0]
                continue

            host[key] = val
            continue

    if host:
        raw_output.append(host)

    return raw_output if raw else _process(raw_output)



# Copyright (c) 2016, Samantha Marshall (http://pewpewthespells.com)
# All rights reserved.
#
# https://github.com/samdmarshall/pbPlist
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of Samantha Marshall nor the names of its contributors may
# be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import string

if sys.version_info >= (3, 0):
    def unichr(character): # pylint: disable=redefined-builtin
        return chr(character)

def ConvertNEXTSTEPToUnicode(hex_digits):
    # taken from http://ftp.unicode.org/Public/MAPPINGS/VENDORS/NEXT/NEXTSTEP.TXT
    conversion = {
        "80":	"a0",	# NO-BREAK SPACE
        "81":	"c0",	# LATIN CAPITAL LETTER A WITH GRAVE
        "82":	"c1",	# LATIN CAPITAL LETTER A WITH ACUTE
        "83":	"c2",	# LATIN CAPITAL LETTER A WITH CIRCUMFLEX
        "84":	"c3",	# LATIN CAPITAL LETTER A WITH TILDE
        "85":	"c4",	# LATIN CAPITAL LETTER A WITH DIAERESIS
        "86":	"c5",	# LATIN CAPITAL LETTER A WITH RING
        "87":	"c7",	# LATIN CAPITAL LETTER C WITH CEDILLA
        "88":	"c8",	# LATIN CAPITAL LETTER E WITH GRAVE
        "89":	"c9",	# LATIN CAPITAL LETTER E WITH ACUTE
        "8a":	"ca",	# LATIN CAPITAL LETTER E WITH CIRCUMFLEX
        "8b":	"cb",	# LATIN CAPITAL LETTER E WITH DIAERESIS
        "8c":	"cc",	# LATIN CAPITAL LETTER I WITH GRAVE
        "8d":	"cd",	# LATIN CAPITAL LETTER I WITH ACUTE
        "8e":	"ce",	# LATIN CAPITAL LETTER I WITH CIRCUMFLEX
        "8f":	"cf",	# LATIN CAPITAL LETTER I WITH DIAERESIS
        "90":	"d0",	# LATIN CAPITAL LETTER ETH
        "91":	"d1",	# LATIN CAPITAL LETTER N WITH TILDE
        "92":	"d2",	# LATIN CAPITAL LETTER O WITH GRAVE
        "93":	"d3",	# LATIN CAPITAL LETTER O WITH ACUTE
        "94":	"d4",	# LATIN CAPITAL LETTER O WITH CIRCUMFLEX
        "95":	"d5",	# LATIN CAPITAL LETTER O WITH TILDE
        "96":	"d6",	# LATIN CAPITAL LETTER O WITH DIAERESIS
        "97":	"d9",	# LATIN CAPITAL LETTER U WITH GRAVE
        "98":	"da",	# LATIN CAPITAL LETTER U WITH ACUTE
        "99":	"db",	# LATIN CAPITAL LETTER U WITH CIRCUMFLEX
        "9a":	"dc",	# LATIN CAPITAL LETTER U WITH DIAERESIS
        "9b":	"dd",	# LATIN CAPITAL LETTER Y WITH ACUTE
        "9c":	"de",	# LATIN CAPITAL LETTER THORN
        "9d":	"b5",	# MICRO SIGN
        "9e":	"d7",	# MULTIPLICATION SIGN
        "9f":	"f7",	# DIVISION SIGN
        "a0":	"a9",	# COPYRIGHT SIGN
        "a1":	"a1",	# INVERTED EXCLAMATION MARK
        "a2":	"a2",	# CENT SIGN
        "a3":	"a3",	# POUND SIGN
        "a4":	"44",	# FRACTION SLASH
        "a5":	"a5",	# YEN SIGN
        "a6":	"92",	# LATIN SMALL LETTER F WITH HOOK
        "a7":	"a7",	# SECTION SIGN
        "a8":	"a4",	# CURRENCY SIGN
        "a9":	"19",	# RIGHT SINGLE QUOTATION MARK
        "aa":	"1c",	# LEFT DOUBLE QUOTATION MARK
        "ab":	"ab",	# LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
        "ac":	"39",	# SINGLE LEFT-POINTING ANGLE QUOTATION MARK
        "ad":	"3a",	# SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
        "ae":	"01",	# LATIN SMALL LIGATURE FI
        "af":	"02",	# LATIN SMALL LIGATURE FL
        "b0":	"ae",	# REGISTERED SIGN
        "b1":	"13",	# EN DASH
        "b2":	"20",	# DAGGER
        "b3":	"21",	# DOUBLE DAGGER
        "b4":	"b7",	# MIDDLE DOT
        "b5":	"a6",	# BROKEN BAR
        "b6":	"b6",	# PILCROW SIGN
        "b7":	"22",	# BULLET
        "b8":	"1a",	# SINGLE LOW-9 QUOTATION MARK
        "b9":	"1e",	# DOUBLE LOW-9 QUOTATION MARK
        "ba":	"1d",	# RIGHT DOUBLE QUOTATION MARK
        "bb":	"bb",	# RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
        "bc":	"26",	# HORIZONTAL ELLIPSIS
        "bd":	"30",	# PER MILLE SIGN
        "be":	"ac",	# NOT SIGN
        "bf":	"bf",	# INVERTED QUESTION MARK
        "c0":	"b9",	# SUPERSCRIPT ONE
        "c1":	"cb",	# MODIFIER LETTER GRAVE ACCENT
        "c2":	"b4",	# ACUTE ACCENT
        "c3":	"c6",	# MODIFIER LETTER CIRCUMFLEX ACCENT
        "c4":	"dc",	# SMALL TILDE
        "c5":	"af",	# MACRON
        "c6":	"d8",	# BREVE
        "c7":	"d9",	# DOT ABOVE
        "c8":	"a8",	# DIAERESIS
        "c9":	"b2",	# SUPERSCRIPT TWO
        "ca":	"da",	# RING ABOVE
        "cb":	"b8",	# CEDILLA
        "cc":	"b3",	# SUPERSCRIPT THREE
        "cd":	"dd",	# DOUBLE ACUTE ACCENT
        "ce":	"db",	# OGONEK
        "cf":	"c7",	# CARON
        "d0":	"14",	# EM DASH
        "d1":	"b1",	# PLUS-MINUS SIGN
        "d2":	"bc",	# VULGAR FRACTION ONE QUARTER
        "d3":	"bd",	# VULGAR FRACTION ONE HALF
        "d4":	"be",	# VULGAR FRACTION THREE QUARTERS
        "d5":	"e0",	# LATIN SMALL LETTER A WITH GRAVE
        "d6":	"e1",	# LATIN SMALL LETTER A WITH ACUTE
        "d7":	"e2",	# LATIN SMALL LETTER A WITH CIRCUMFLEX
        "d8":	"e3",	# LATIN SMALL LETTER A WITH TILDE
        "d9":	"e4",	# LATIN SMALL LETTER A WITH DIAERESIS
        "da":	"e5",	# LATIN SMALL LETTER A WITH RING ABOVE
        "db":	"e7",	# LATIN SMALL LETTER C WITH CEDILLA
        "dc":	"e8",	# LATIN SMALL LETTER E WITH GRAVE
        "dd":	"e9",	# LATIN SMALL LETTER E WITH ACUTE
        "de":	"ea",	# LATIN SMALL LETTER E WITH CIRCUMFLEX
        "df":	"eb",	# LATIN SMALL LETTER E WITH DIAERESIS
        "e0":	"ec",	# LATIN SMALL LETTER I WITH GRAVE
        "e1":	"c6",	# LATIN CAPITAL LETTER AE
        "e2":	"ed",	# LATIN SMALL LETTER I WITH ACUTE
        "e3":	"aa",	# FEMININE ORDINAL INDICATOR
        "e4":	"ee",	# LATIN SMALL LETTER I WITH CIRCUMFLEX
        "e5":	"ef",	# LATIN SMALL LETTER I WITH DIAERESIS
        "e6":	"f0",	# LATIN SMALL LETTER ETH
        "e7":	"f1",	# LATIN SMALL LETTER N WITH TILDE
        "e8":	"41",	# LATIN CAPITAL LETTER L WITH STROKE
        "e9":	"d8",	# LATIN CAPITAL LETTER O WITH STROKE
        "ea":	"52",	# LATIN CAPITAL LIGATURE OE
        "eb":	"ba",	# MASCULINE ORDINAL INDICATOR
        "ec":	"f2",	# LATIN SMALL LETTER O WITH GRAVE
        "ed":	"f3",	# LATIN SMALL LETTER O WITH ACUTE
        "ee":	"f4",	# LATIN SMALL LETTER O WITH CIRCUMFLEX
        "ef":	"f5",	# LATIN SMALL LETTER O WITH TILDE
        "f0":	"f6",	# LATIN SMALL LETTER O WITH DIAERESIS
        "f1":	"e6",	# LATIN SMALL LETTER AE
        "f2":	"f9",	# LATIN SMALL LETTER U WITH GRAVE
        "f3":	"fa",	# LATIN SMALL LETTER U WITH ACUTE
        "f4":	"fb",	# LATIN SMALL LETTER U WITH CIRCUMFLEX
        "f5":	"31",	# LATIN SMALL LETTER DOTLESS I
        "f6":	"fc",	# LATIN SMALL LETTER U WITH DIAERESIS
        "f7":	"fd",	# LATIN SMALL LETTER Y WITH ACUTE
        "f8":	"42",	# LATIN SMALL LETTER L WITH STROKE
        "f9":	"f8",	# LATIN SMALL LETTER O WITH STROKE
        "fa":	"53",	# LATIN SMALL LIGATURE OE
        "fb":	"df",	# LATIN SMALL LETTER SHARP S
        "fc":	"fe",	# LATIN SMALL LETTER THORN
        "fd":	"ff",	# LATIN SMALL LETTER Y WITH DIAERESIS
        "fe":	"fd",	# .notdef, REPLACEMENT CHARACTER
        "ff":	"fd",	# .notdef, REPLACEMENT CHARACTER
    }
    return conversion[hex_digits]

def IsOctalNumber(character):
    oct_digits = set(string.octdigits)
    return set(character).issubset(oct_digits)

def IsHexNumber(character):
    hex_digits = set(string.hexdigits)
    return set(character).issubset(hex_digits)

def SanitizeCharacter(character):
    char = character
    escaped_characters = {
        '\a': '\\a',
        '\b': '\\b',
        '\f': '\\f',
        '\n': '\\n',
        '\r': '\\r',
        '\t': '\\t',
        '\v': '\\v',
        '\"': '\\"',
    }
    if character in escaped_characters.keys():
        char = escaped_characters[character]
    return char

# http://www.opensource.apple.com/source/CF/CF-744.19/CFOldStylePList.c See `getSlashedChar()`
def UnQuotifyString(string_data, start_index, end_index): # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    formatted_string = ''
    extracted_string = string_data[start_index:end_index]
    string_length = len(extracted_string)
    all_cases = ['0', '1', '2', '3', '4', '5', '6', '7', 'a', 'b', 'f', 'n', 'r', 't', 'v', '\"', '\n', 'U']
    index = 0
    while index < string_length: # pylint: disable=too-many-nested-blocks
        current_char = extracted_string[index]
        if current_char == '\\':
            next_char = extracted_string[index+1]
            if next_char in all_cases:
                index += 1
                if next_char == 'a':
                    formatted_string += '\a'
                if next_char == 'b':
                    formatted_string += '\b'
                if next_char == 'f':
                    formatted_string += '\f'
                if next_char == 'n':
                    formatted_string += '\n'
                if next_char == 'r':
                    formatted_string += '\r'
                if next_char == 't':
                    formatted_string += '\t'
                if next_char == 'v':
                    formatted_string += '\v'
                if next_char == '"':
                    formatted_string += '\"'
                if next_char == '\n':
                    formatted_string += '\n'
                if next_char == 'U':
                    starting_index = index + 1
                    ending_index = starting_index + 4
                    unicode_numbers = extracted_string[starting_index:ending_index]
                    for number in unicode_numbers:
                        index += 1
                        if IsHexNumber(number) is False: # pragma: no cover
                            message = 'Invalid unicode sequence on line '+str(LineNumberForIndex(string_data, start_index+index))
                            raise Exception(message)
                    formatted_string += unichr(int(unicode_numbers, 16))
                if IsOctalNumber(next_char) is True: # https://twitter.com/Catfish_Man/status/658014170055507968
                    starting_index = index
                    ending_index = starting_index + 1
                    for oct_index in range(3):
                        test_index = starting_index + oct_index
                        test_oct = extracted_string[test_index]
                        if IsOctalNumber(test_oct) is True:
                            ending_index += 1
                    octal_numbers = extracted_string[starting_index:ending_index]
                    hex_number = int(octal_numbers, 8)
                    hex_str = format(hex_number, 'x')
                    if hex_number >= 0x80:
                        hex_str = ConvertNEXTSTEPToUnicode(hex_str)
                    formatted_string += unichr(int('00'+hex_str, 16))
            else:
                formatted_string += current_char
                index += 1
                formatted_string += next_char
        else:
            formatted_string += current_char
        index += 1
    return formatted_string

def LineNumberForIndex(string_data, current_index):
    line_number = 1
    index = 0
    string_length = len(string_data)
    while (index < current_index) and (index < string_length):
        current_char = string_data[index]
        if IsNewline(current_char) is True:
            line_number += 1
        index += 1
    return line_number

def IsValidUnquotedStringCharacter(character):
    if len(character) == 1:
        valid_characters = set(string.ascii_letters+string.digits+'_$/:.-')
        return set(character).issubset(valid_characters)
    else: # pragma: no cover
        message = 'The function "IsValidUnquotedStringCharacter()" can only take single characters!'
        raise ValueError(message)

def IsSpecialWhitespace(character):
    value = ord(character)
    result = (value >= 9 and value <= 13) # tab, newline, vt, form feed, carriage return
    return result

def IsUnicodeSeparator(character):
    value = ord(character)
    result = (value == 8232 or value == 8233)
    return result

def IsRegularWhitespace(character):
    value = ord(character)
    result = (value == 32 or IsUnicodeSeparator(character)) # space and Unicode line sep, para sep
    return result

def IsDataFormattingWhitespace(character):
    value = ord(character)
    result = (IsNewline(character) or IsRegularWhitespace(character) or value == 9)
    return result

def IsNewline(character):
    value = ord(character)
    result = (value == 13 or value == 10)
    return result

def IsEndOfLine(character):
    result = (IsNewline(character) or IsUnicodeSeparator(character))
    return result

def IndexOfNextNonSpace(string_data, current_index): # pylint: disable=too-many-branches,too-many-statements
    successful = False
    found_index = current_index
    string_length = len(string_data)
    annotation_string = ''
    while found_index < string_length: # pylint: disable=too-many-nested-blocks
        current_char = string_data[found_index]
        if IsSpecialWhitespace(current_char) is True:
            found_index += 1
            continue
        if IsRegularWhitespace(current_char) is True:
            found_index += 1
            continue
        if current_char == '/':
            next_index = found_index + 1
            if next_index >= string_length:
                successful = True
                break
            else:
                next_character = string_data[next_index]
                if next_character == '/': # found a line comment "//"
                    found_index += 1
                    next_index = found_index
                    first_pass = True
                    while next_index < string_length:
                        test_char = string_data[next_index]
                        if IsEndOfLine(test_char) is True:
                            break
                        else:
                            if first_pass is False:
                                annotation_string += test_char
                            else:
                                first_pass = False
                        next_index += 1
                    found_index = next_index
                elif next_character == '*': # found a block comment "/* ... */"
                    found_index += 1
                    next_index = found_index
                    first_pass = True
                    while next_index < string_length:
                        test_char = string_data[next_index]
                        if test_char == '*' and (next_index+1 < string_length) and string_data[next_index+1] == '/':
                            next_index += 2
                            break
                        else:
                            if first_pass != True:
                                annotation_string += test_char
                            else:
                                first_pass = False
                        next_index += 1
                    found_index = next_index
                else:
                    successful = True
                    break
        else:
            successful = True
            break
    result = (successful, found_index, annotation_string)
    return result



"""jc - JSON Convert `pip-show` command output parser

Usage (cli):

    $ pip show | jc --pip-show

or

    $ jc pip show

Usage (module):

    import jc
    result = jc.parse('pip_show', pip_show_command_output)

Schema:

    [
      {
        "name":             string,
        "version":          string,
        "summary":          string,
        "home_page":        string,
        "author":           string,
        "author_email":     string,
        "license":          string,
        "location":         string,
        "requires":         string,
        "required_by":      string
      }
    ]

Examples:

    $ pip show wrapt jc wheel | jc --pip-show -p
    [
      {
        "name": "wrapt",
        "version": "1.11.2",
        "summary": "Module for decorators, wrappers and monkey patching.",
        "home_page": "https://github.com/GrahamDumpleton/wrapt",
        "author": "Graham Dumpleton",
        "author_email": "Graham.Dumpleton@gmail.com",
        "license": "BSD",
        "location": "/usr/local/lib/python3.7/site-packages",
        "requires": null,
        "required_by": "astroid"
      },
      {
        "name": "wheel",
        "version": "0.33.4",
        "summary": "A built-package format for Python.",
        "home_page": "https://github.com/pypa/wheel",
        "author": "Daniel Holth",
        "author_email": "dholth@fastmail.fm",
        "license": "MIT",
        "location": "/usr/local/lib/python3.7/site-packages",
        "requires": null,
        "required_by": null
      }
    ]
"""
from typing import List, Dict, Optional
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.4'
    description = '`pip show` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    magic_commands = ['pip show', 'pip3 show']
    tags = ['command']


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """
    # no further processing
    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []
    package: Dict = {}
    last_key: str = ''
    last_key_data: List = []

    # Clear any blank lines
    cleandata = list(filter(None, data.splitlines()))

    if jc.utils.has_data(data):

        for row in cleandata:
            if row.startswith('---'):
                if last_key_data:
                    package[last_key] = package[last_key] + '\n' + '\n'.join(last_key_data)

                raw_output.append(package)
                package = {}
                last_key = ''
                last_key_data = []
                continue

            if not row.startswith(' '):
                item_key = row.split(': ', maxsplit=1)[0].lower().replace('-', '_')
                item_value: Optional[str] = row.split(': ', maxsplit=1)[1]

                if item_value == '':
                    item_value = None

                if last_key_data and last_key != item_key:
                    package[last_key] = package[last_key] + '\n' + '\n'.join(last_key_data)
                    last_key_data = []

                package[item_key] = item_value
                last_key = item_key
                continue

            if row.startswith(' '):
                last_key_data.append(row.strip())
                continue

        if package:
            if last_key_data:
                package[last_key] = package[last_key] + '\n' + '\n'.join(last_key_data)

            raw_output.append(package)

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert `/usr/bin/time` command output parser

Output from `/usr/bin/time` is sent to `STDERR`, so the `-o` option can be
used to redirect the output to a file that can be read by `jc`.

Alternatively, the output from `/usr/bin/time` can be redirected to `STDOUT`
so `jc` can receive it.

> Note: `/usr/bin/time` is similar but different from the Bash builtin
> `time` command.

Usage (cli):

    $ /usr/bin/time -o timefile.out sleep 2; cat timefile.out | \\
      jc --time -p

Usage (module):

    import jc
    result = jc.parse('time', time_command_output)

Schema:

Source: https://www.freebsd.org/cgi/man.cgi?query=getrusage,
https://man7.org/linux/man-pages/man1/time.1.html

    {
      "real_time":                          float,
      "user_time":                          float,
      "system_time":                        float,
      "elapsed_time":                       string,
      "elapsed_time_hours":                 integer,
      "elapsed_time_minutes":               integer,
      "elapsed_time_seconds":               integer,
      "elapsed_time_centiseconds":          integer,
      "elapsed_time_total_seconds":         float,
      "cpu_percent":                        integer,   # null if ?
      "average_shared_text_size":           integer,
      "average_unshared_data_size":         integer,
      "average_unshared_stack_size":        integer,
      "average_shared_memory_size":         integer,
      "maximum_resident_set_size":          integer,
      "block_input_operations":             integer,   # [0]
      "block_output_operations":            integer,   # [1]
      "major_pagefaults":                   integer,
      "minor_pagefaults":                   integer,
      "swaps":                              integer,
      "page_reclaims":                      integer,
      "page_faults":                        integer,
      "messages_sent":                      integer,
      "messages_received":                  integer,
      "signals_received":                   integer,
      "voluntary_context_switches":         integer,
      "involuntary_context_switches":       integer
      "command_being_timed":                string,
      "average_stack_size":                 integer,
      "average_total_size":                 integer,
      "average_resident_set_size":          integer,
      "signals_delivered":                  integer,
      "page_size":                          integer,
      "exit_status":                        integer
    }

    [0] aka File system inputs
    [1] aka File system outputs

Examples:

    $ /usr/bin/time --verbose -o timefile.out sleep 2; cat timefile.out | \\
      jc --time -p
    {
      "command_being_timed": "sleep 2",
      "user_time": 0.0,
      "system_time": 0.0,
      "cpu_percent": 0,
      "elapsed_time": "0:02.00",
      "average_shared_text_size": 0,
      "average_unshared_data_size": 0,
      "average_stack_size": 0,
      "average_total_size": 0,
      "maximum_resident_set_size": 2084,
      "average_resident_set_size": 0,
      "major_pagefaults": 0,
      "minor_pagefaults": 72,
      "voluntary_context_switches": 2,
      "involuntary_context_switches": 1,
      "swaps": 0,
      "block_input_operations": 0,
      "block_output_operations": 0,
      "messages_sent": 0,
      "messages_received": 0,
      "signals_delivered": 0,
      "page_size": 4096,
      "exit_status": 0,
      "elapsed_time_hours": 0,
      "elapsed_time_minutes": 0,
      "elapsed_time_seconds": 2,
      "elapsed_time_centiseconds": 50,
      "elapsed_time_total_seconds": 2.5
    }

    $ /usr/bin/time --verbose -o timefile.out sleep 2; cat timefile.out | \\
      jc --time -p -r
    {
      "command_being_timed": "\"sleep 2\"",
      "user_time": "0.00",
      "system_time": "0.00",
      "cpu_percent": "0",
      "elapsed_time": "0:02.00",
      "average_shared_text_size": "0",
      "average_unshared_data_size": "0",
      "average_stack_size": "0",
      "average_total_size": "0",
      "maximum_resident_set_size": "2084",
      "average_resident_set_size": "0",
      "major_pagefaults": "0",
      "minor_pagefaults": "72",
      "voluntary_context_switches": "2",
      "involuntary_context_switches": "0",
      "swaps": "0",
      "block_input_operations": "0",
      "block_output_operations": "0",
      "messages_sent": "0",
      "messages_received": "0",
      "signals_delivered": "0",
      "page_size": "4096",
      "exit_status": "0"
    }
"""
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.4'
    description = '`/usr/bin/time` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'cygwin', 'aix', 'freebsd']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        Dictionary. Structured data to conform to the schema.
    """
    int_list = {
        'cpu_percent', 'average_shared_text_size', 'average_unshared_data_size',
        'average_unshared_stack_size', 'average_shared_memory_size', 'maximum_resident_set_size',
        'block_input_operations', 'block_output_operations', 'major_pagefaults', 'minor_pagefaults',
        'swaps', 'page_reclaims', 'page_faults', 'messages_sent', 'messages_received',
        'signals_received', 'voluntary_context_switches', 'involuntary_context_switches',
        'average_stack_size', 'average_total_size', 'average_resident_set_size',
        'signals_delivered', 'page_size', 'exit_status'
    }

    float_list = {'real_time', 'user_time', 'system_time'}

    if 'command_being_timed' in proc_data:
        proc_data['command_being_timed'] = proc_data['command_being_timed'][1:-1]

    if 'elapsed_time' in proc_data:
        proc_data['elapsed_time'] = proc_data['elapsed_time'].replace('.', ':')
        *hours, minutes, seconds, centiseconds = proc_data['elapsed_time'].split(':')
        proc_data['elapsed_time'] = proc_data['elapsed_time'][::-1].replace(':', '.', 1)[::-1]
        if hours:
            proc_data['elapsed_time_hours'] = jc.utils.convert_to_int(hours[0])
        else:
            proc_data['elapsed_time_hours'] = 0
        proc_data['elapsed_time_minutes'] = jc.utils.convert_to_int(minutes)
        proc_data['elapsed_time_seconds'] = jc.utils.convert_to_int(seconds)
        proc_data['elapsed_time_centiseconds'] = jc.utils.convert_to_int(centiseconds)
        proc_data['elapsed_time_total_seconds'] = (proc_data['elapsed_time_hours'] * 3600) + \
                                                  (proc_data['elapsed_time_minutes'] * 60) + \
                                                  (proc_data['elapsed_time_seconds']) + \
                                                  (proc_data['elapsed_time_centiseconds'] / 100)

    # convert ints and floats
    for key in proc_data:
        if key in int_list:
            proc_data[key] = jc.utils.convert_to_int(proc_data[key])

        if key in float_list:
            proc_data[key] = jc.utils.convert_to_float(proc_data[key])

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = {}

    if jc.utils.has_data(data):
        time_type = None     # linux_brief, linux_long, bsd_brief, bsd_long, posix

        for line in filter(None, data.splitlines()):
            # linux default style:
            # 0.00user 0.00system 0:03.00elapsed 0%CPU (0avgtext+0avgdata 2148maxresident)k
            # 0inputs+0outputs (0major+71minor)pagefaults 0swaps
            if time_type != 'linux_brief' and 'elapsed' in line:
                time_type = 'linux_brief'

            # BSD/OSX default style:
            #         0.00 real         0.00 user         0.00 sys
            elif time_type != 'bsd_brief' and ' user ' in line:
                time_type = 'bsd_brief'

            elif time_type != 'linux_long' and 'Command' in line:
                time_type = 'linux_long'

            elif time_type != 'bsd_long' and 'maximum resident set size' in line:
                time_type = 'bsd_long'

            # POSIX compliant output:
            # real         0.00
            # user         0.00
            # sys          0.00
            elif time_type != 'posix' and line.startswith('real '):
                time_type = 'posix'

            # start parsing lines
            if time_type == 'linux_brief':
                if 'elapsed' in line:
                    line_num = 0
                else:
                    line_num = 1

                new_line = line.replace('+', ' ').replace('(', ' ').replace(')', ' ')\
                               .replace('user', ' ').replace('system', ' ').replace('elapsed', ' ')\
                               .replace('elapsed', ' ').replace('%CPU', ' ').replace('avgtext', ' ')\
                               .replace('avgdata', ' ').replace('maxresident', ' ').replace('inputs', ' ')\
                               .replace('outputs', ' ').replace('major', ' ').replace('minor', ' ')\
                               .replace('pagefaults', ' ').replace('swaps', ' ').replace('k', ' ')

                linux_brief_line = new_line.split()

                if line_num == 0:
                    raw_output['user_time'] = linux_brief_line[0]
                    raw_output['system_time'] = linux_brief_line[1]
                    raw_output['elapsed_time'] = linux_brief_line[2]
                    raw_output['cpu_percent'] = None if linux_brief_line[3] == '?' else linux_brief_line[3]
                    raw_output['average_shared_text'] = linux_brief_line[4]
                    raw_output['average_unshared_data_size'] = linux_brief_line[5]
                    raw_output['maximum_resident_set_size'] = linux_brief_line[6]
                else:
                    raw_output['block_input_operations'] = linux_brief_line[0]
                    raw_output['block_output_operations'] = linux_brief_line[1]
                    raw_output['major_pagefaults'] = linux_brief_line[2]
                    raw_output['minor_pagefaults'] = linux_brief_line[3]
                    raw_output['swaps'] = linux_brief_line[4]

            if time_type == 'posix':
                posix_line = line.split()
                if 'real' in line:
                    raw_output['real_time'] = posix_line[1]
                if 'user' in line:
                    raw_output['user_time'] = posix_line[1]
                if 'sys' in line:
                    raw_output['system_time'] = posix_line[1]

            if time_type == 'bsd_brief':
                bsd_brief_line = line.split()
                raw_output['real_time'] = bsd_brief_line[0]
                raw_output['user_time'] = bsd_brief_line[2]
                raw_output['system_time'] = bsd_brief_line[4]

            if time_type == 'bsd_long':
                bsd_long_line = line.split(maxsplit=1)
                key = bsd_long_line[1].replace(' ', '_')

                # fixup some key names
                if key == 'average_shared_text':
                    key = 'average_shared_text_size'

                value = bsd_long_line[0]
                raw_output[key] = value

            if time_type == 'linux_long':
                # cleanup key names: (h:mm:ss or m:ss)
                # line = line.replace('h:mm:ss', '', 1).replace('m:ss', '')
                linux_long_line = line.split(': ', maxsplit=1)
                key = linux_long_line[0].strip().lower().replace(' ', '_').replace('(', '').replace(')', '')\
                                        .replace('/', '_').replace(':', '_').replace('_kbytes', '')\
                                        .replace('_seconds', '').replace('socket_', '').replace('_bytes', '')

                # fixup some key names
                if key == 'file_system_inputs':
                    key = 'block_input_operations'

                if key == 'file_system_outputs':
                    key = 'block_output_operations'

                if key == 'percent_of_cpu_this_job_got':
                    key = 'cpu_percent'

                if key == 'elapsed_wall_clock_time_h_mm_ss_or_m_ss':
                    key = 'elapsed_time'

                if key == 'major_requiring_i_o_page_faults':
                    key = 'major_pagefaults'

                if key == 'minor_reclaiming_a_frame_page_faults':
                    key = 'minor_pagefaults'

                value = linux_long_line[1].replace('%', '')
                raw_output[key] = value

    if raw:
        return raw_output
    else:
        return _process(raw_output)



"""jc - JSON Convert `csv` file streaming parser

> This streaming parser outputs JSON Lines (cli) or returns an Iterable of
> Dictionaries (module)

The `csv` streaming parser will attempt to automatically detect the
delimiter character. If the delimiter cannot be detected it will default
to comma. The first row of the file must be a header row.

> Note: The first 100 rows are read into memory to enable delimiter
> detection, then the rest of the rows are loaded lazily.

Usage (cli):

    $ cat file.csv | jc --csv-s

Usage (module):

    import jc

    result = jc.parse('csv_s', csv_output.splitlines())
    for item in result:
        # do something

Schema:

CSV file converted to a Dictionary:
https://docs.python.org/3/library/csv.html

    {
      "column_name1":     string,
      "column_name2":     string,

      # below object only exists if using -qq or ignore_exceptions=True
      "_jc_meta": {
        "success":        boolean,     # false if error parsing
        "error":          string,      # exists if "success" is false
        "line":           string       # exists if "success" is false
      }
    }

Examples:

    $ cat homes.csv
    "Sell", "List", "Living", "Rooms", "Beds", "Baths", "Age", "Acres"...
    142, 160, 28, 10, 5, 3,  60, 0.28,  3167
    175, 180, 18,  8, 4, 1,  12, 0.43,  4033
    129, 132, 13,  6, 3, 1,  41, 0.33,  1471
    ...

    $ cat homes.csv | jc --csv-s
    {"Sell":"142","List":"160","Living":"28","Rooms":"10","Beds":"5"...}
    {"Sell":"175","List":"180","Living":"18","Rooms":"8","Beds":"4"...}
    {"Sell":"129","List":"132","Living":"13","Rooms":"6","Beds":"3"...}
    ...
"""
import itertools
import csv
import jc.utils
from jc.streaming import streaming_input_type_check, add_jc_meta, raise_or_yield
from jc.exceptions import ParseError


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.4'
    description = 'CSV file streaming parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    details = 'Using the python standard csv library'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['standard', 'file', 'string']
    streaming = True


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Each Dictionary represents a row in the csv
        file.
    """
    # No further processing
    return proc_data


@add_jc_meta
def parse(data, raw=False, quiet=False, ignore_exceptions=False):
    """
    Main text parsing generator function. Returns an iterable object.

    Parameters:

        data:              (iterable)  line-based text data to parse
                                       (e.g. sys.stdin or str.splitlines())

        raw:               (boolean)   unprocessed output if True
        quiet:             (boolean)   suppress warning messages if True
        ignore_exceptions: (boolean)   ignore parsing exceptions if True

    Returns:

        Iterable of Dictionaries
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    streaming_input_type_check(data)

    # convert data to an iterable in case a sequence like a list is used as input.
    # this allows the exhaustion of the input so we don't double-process later.
    data = iter(data)
    temp_list = []

    # first, load the first 100 lines into a list to detect the CSV dialect
    for line in itertools.islice(data, 100):
        temp_list.append(line.rstrip())

    # check for Python bug that does not split on `\r` newlines from sys.stdin correctly
    # https://bugs.python.org/issue45617
    if len(temp_list) == 1:
        raise ParseError('Unable to detect line endings. Please try the non-streaming CSV parser instead.')

    # remove BOM bytes from first row, if present
    if temp_list:
        if isinstance(temp_list[0], str):
            temp_list[0] = temp_list[0].encode('utf-8')

        temp_list[0] = temp_list[0].decode('utf-8-sig')

    sniffdata = '\r\n'.join(temp_list)[:1024]
    dialect = 'excel'  # default in csv module

    try:
        dialect = csv.Sniffer().sniff(sniffdata)
        if '""' in sniffdata:
            dialect.doublequote = True
    except Exception:
        pass

    # chain `temp_list` and `data` together to lazy load the rest of the CSV data
    new_data = itertools.chain(temp_list, data)
    reader = csv.DictReader(new_data, dialect=dialect)

    for row in reader:
        try:
            yield row if raw else _process(row)
        except Exception as e:
            yield raise_or_yield(ignore_exceptions, e, str(row))



"""jc - JSON Convert Version string output parser

Best-effort attempt to parse various styles of version numbers. This parser
is based off of the version parser included in the CPython distutils
libary.

If the version string conforms to some de facto-standard versioning rules
followed by many developers a `strict` key will be present in the output
with a value of `true` along with the named parsed components.

All other version strings will have a `strict` value of `false` and a
`components` key will contain a list of detected parts of the version
string.

See Also: `semver` parser.

Usage (cli):

    $ echo 1.2a1 | jc --ver

Usage (module):

    import jc
    result = jc.parse('ver', version_string_output)

Schema:

    {
      "major":                  integer,
      "minor":                  integer,
      "patch":                  integer,
      "prerelease":             string,
      "prerelease_num":         integer,
      "components": [
                                integer/string
      ],
      "strict":                 boolean
    }

Examples:

    $ echo 1.2a1 | jc --ver -p
    {
      "major": 1,
      "minor": 2,
      "patch": 0,
      "prerelease": "a",
      "prerelease_num": 1,
      "strict": true
    }

    $ echo 1.2a1 | jc --ver -p -r
    {
      "major": "1",
      "minor": "2",
      "patch": "0",
      "prerelease": "a",
      "prerelease_num": "1",
      "strict": true
    }

    $ echo 1.2beta3 | jc --ver -p
    {
      "components": [
        1,
        2,
        "beta",
        3
      ],
      "strict": false
    }

    $ echo 1.2beta3 | jc --ver -p -r
    {
      "components": [
        "1",
        "2",
        "beta",
        "3"
      ],
      "strict": false
    }
"""
import re
from typing import Dict
from jc.jc_types import JSONDictType
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = 'Version string parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    details = 'Based on distutils/version.py from CPython 3.9.5.'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['generic', 'string']


__version__ = info.version


def _process(proc_data: JSONDictType) -> JSONDictType:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'major', 'minor', 'patch', 'prerelease', 'prerelease_num'}

    for k, v in proc_data.items():
        if k in int_list:
            try:
                proc_data[k] = int(v)
            except Exception:
                pass

    if 'components' in proc_data:
        for i, obj in enumerate(proc_data['components']):
            try:
                proc_data['components'][i] = int(obj)
            except Exception:
                pass

    return proc_data


def strict_parse(vstring):
    version_re = re.compile(r'^(\d+) \. (\d+) (\. (\d+))? ([ab](\d+))?$', re.VERBOSE)
    match = version_re.match(vstring)
    if not match:
        raise ValueError("invalid version number '%s'" % vstring)

    (major, minor, patch, prerelease, prerelease_num) = \
        match.group(1, 2, 4, 5, 6)

    if not patch:
        patch = '0'

    if prerelease:
        prerelease = prerelease[0]
    else:
        prerelease = None

    return {
        'major': major,
        'minor': minor,
        'patch': patch,
        'prerelease': prerelease,
        'prerelease_num': prerelease_num
    }


def loose_parse(vstring):
    component_re = re.compile(r'(\d+ | [a-z]+ | \.)', re.VERBOSE)
    components = [x for x in component_re.split(vstring) if x and x != '.']

    return components


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> JSONDictType:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: Dict = {}
    strict = True

    if jc.utils.has_data(data):

        # based on distutils/version.py from CPython 3.9.5
        # PSF License (see https://opensource.org/licenses/Python-2.0)

        data = data.strip()

        try:
            raw_output = strict_parse(data)

        except ValueError:
            raw_output['components'] = loose_parse(data)
            strict = False

        if raw_output:
            raw_output['strict'] = strict

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert Syslog RFC 3164 string parser

This parser accepts a single syslog line string or multiple syslog lines
separated by newlines. A warning message to `STDERR` will be printed if an
unparsable line is found unless `--quiet` or `quiet=True` is used.

Usage (cli):

    $ echo '<34>Oct 11 22:14:15 mymachine su: su root...' | jc --syslog-bsd

Usage (module):

    import jc
    result = jc.parse('syslog_bsd', syslog_command_output)

Schema:

    [
      {
        "priority":                   integer/null,
        "date":                       string,
        "hostname":                   string,
        "tag":                        string/null,
        "content":                    string,
        "unparsable":                 string,  # [0]
      }
    ]

    [0] this field exists if the syslog line is not parsable. The value
        is the original syslog line.

Examples:

    $ cat syslog.txt | jc --syslog-bsd -p
    [
      {
        "priority": 34,
        "date": "Oct 11 22:14:15",
        "hostname": "mymachine",
        "tag": "su",
        "content": "'su root' failed for lonvick on /dev/pts/8"
      }
    ]

    $ cat syslog.txt | jc --syslog-bsd -p -r
    [
      {
        "priority": "34",
        "date": "Oct 11 22:14:15",
        "hostname": "mymachine",
        "tag": "su",
        "content": "'su root' failed for lonvick on /dev/pts/8"
      }
    ]
"""
import re
from typing import List, Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = 'Syslog RFC 3164 string parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['standard', 'file', 'string']


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'priority'}

    for item in proc_data:
        for key in item:
                if key in int_list:
                    item[key] = jc.utils.convert_to_int(item[key])

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []
    syslog_dict = {}

    # inspired by https://gist.github.com/miticojo/b16bb13e78572c2d2fac82d9516d5c32
    syslog = re.compile(r'''
        (?P<priority><\d{1,3}>)?
        (?P<header>
            (?P<date>[A-Z][a-z][a-z]\s{1,2}\d{1,2}\s\d{2}?:\d{2}:\d{2})?\s
            (?P<host>[\w][\w\d\.:@-]*)?\s
        )
        (?P<msg>
            (?P<tag>\w+)?
            (?P<content>.*)
        )
        ''', re.VERBOSE
    )

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):
            syslog_match = syslog.match(line)
            if syslog_match:
                priority = None
                if syslog_match.group('priority'):
                    priority = syslog_match.group('priority')[1:-1]

                # check for missing tag
                hostname = syslog_match.group('host')
                tag = syslog_match.group('tag')
                content = syslog_match.group('content')
                if hostname:
                    if hostname.endswith(':'):
                        content = tag + content
                        tag = None
                        hostname = hostname[:-1]

                syslog_dict = {
                    'priority': priority,
                    'date': syslog_match.group('date'),
                    'hostname': hostname,
                    # 'raw_msg': syslog_match.group('msg'),
                    'tag': tag,
                    'content': content.lstrip(' :').rstrip()
                }

            else:
                syslog_dict = {
                    'unparsable': line
                }

                if not quiet:
                    jc.utils.warning_message(
                        [f'Unparsable line found: {line}']
                    )

            if syslog_dict:
                raw_output.append(syslog_dict)

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert `/proc/modules` file parser

Usage (cli):

    $ cat /proc/modules | jc --proc

or

    $ jc /proc/modules

or

    $ cat /proc/modules | jc --proc-modules

Usage (module):

    import jc
    result = jc.parse('proc', proc_modules_file)

or

    import jc
    result = jc.parse('proc_modules', proc_modules_file)

Schema:

    [
      {
        "module":                   string,
        "size":                     integer,
        "used":                     integer,
        "used_by": [
                                    string
        ],
        "status":                   string,
        "location":                 string
      }
    ]

Examples:

    $ cat /proc/modules | jc --proc -p
    [
      {
        "module": "binfmt_misc",
        "size": 24576,
        "used": 1,
        "used_by": [],
        "status": "Live",
        "location": "0xffffffffc0ab4000"
      },
      {
        "module": "vsock_loopback",
        "size": 16384,
        "used": 0,
        "used_by": [],
        "status": "Live",
        "location": "0xffffffffc0a14000"
      },
      {
        "module": "vmw_vsock_virtio_transport_common",
        "size": 36864,
        "used": 1,
        "used_by": [
          "vsock_loopback"
        ],
        "status": "Live",
        "location": "0xffffffffc0a03000"
      },
      ...
    ]

    $ cat /proc/modules | jc --proc-modules -p -r
    [
      {
        "module": "binfmt_misc",
        "size": "24576",
        "used": "1",
        "used_by": [],
        "status": "Live",
        "location": "0xffffffffc0ab4000"
      },
      {
        "module": "vsock_loopback",
        "size": "16384",
        "used": "0",
        "used_by": [],
        "status": "Live",
        "location": "0xffffffffc0a14000"
      },
      {
        "module": "vmw_vsock_virtio_transport_common",
        "size": "36864",
        "used": "1",
        "used_by": [
          "vsock_loopback"
        ],
        "status": "Live",
        "location": "0xffffffffc0a03000"
      },
      ...
    ]
"""
from typing import List, Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/modules` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'size', 'used'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):

            module, size, used, used_by, status, location = line.split()
            used_by_list = used_by.split(',')[:-1]

            raw_output.append(
                {
                    'module': module,
                    'size': size,
                    'used': used,
                    'used_by': used_by_list,
                    'status': status,
                    'location': location
                }
            )

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert `Key/Value` file and string parser

Supports files containing simple key/value pairs.

- Delimiter can be `=` or `:`. Missing values are supported.
- Comment prefix can be `#` or `;`. Comments must be on their own line.
- If duplicate keys are found, only the last value will be used.

> Note: Values starting and ending with quotation marks will have the marks
> removed. If you would like to keep the quotation marks, use the `-r`
> command-line argument or the `raw=True` argument in `parse()`.

Usage (cli):

    $ cat foo.txt | jc --kv

Usage (module):

    import jc
    result = jc.parse('kv', kv_file_output)

Schema:

Key/Value document converted to a dictionary - see the python configparser
standard library documentation for more details.

    {
      "key1":       string,
      "key2":       string
    }

Examples:

    $ cat keyvalue.txt
    # this file contains key/value pairs
    name = John Doe
    address=555 California Drive
    age: 34

    ; comments can include # or ;
    # delimiter can be = or :
    # quoted values have quotation marks stripped by default
    # but can be preserved with the -r argument
    occupation:"Engineer"

    $ cat keyvalue.txt | jc --kv -p
    {
      "name": "John Doe",
      "address": "555 California Drive",
      "age": "34",
      "occupation": "Engineer"
    }
"""
import jc.utils
import configparser


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '2.0'
    description = 'Key/Value file and string parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    details = 'Using configparser from the python standard library'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['generic', 'file', 'string']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary representing a Key/Value pair document.
    """
    # remove quotation marks from beginning and end of values
    for key in proc_data:
        if proc_data[key] is None:
            proc_data[key] = ''

        elif proc_data[key].startswith('"') and proc_data[key].endswith('"'):
            proc_data[key] = proc_data[key][1:-1]

        elif proc_data[key].startswith("'") and proc_data[key].endswith("'"):
            proc_data[key] = proc_data[key][1:-1]

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary representing a Key/Value pair document.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = {}

    if jc.utils.has_data(data):

        kv_parser = configparser.ConfigParser(
            allow_no_value=True,
            interpolation=None,
            default_section=None,
            strict=False
        )

        # don't convert keys to lower-case:
        kv_parser.optionxform = lambda option: option

        data = '[data]\n' + data
        kv_parser.read_string(data)
        output_dict = {s: dict(kv_parser.items(s)) for s in kv_parser.sections()}
        for key, value in output_dict['data'].items():
            raw_output[key] = value

    return raw_output if raw else _process(raw_output)




"""jc - JSON Convert `/proc/version` file parser

> Note: This parser will parse `/proc/version` files that follow the
> common format used by most popular linux distributions.

Usage (cli):

    $ cat /proc/version | jc --proc

or

    $ jc /proc/version

or

    $ cat /proc/version | jc --proc-version

Usage (module):

    import jc
    result = jc.parse('proc', proc_version_file)

or

    import jc
    result = jc.parse('proc_version', proc_version_file)

Schema:

    {
      "version":                  string,
      "email":                    string,
      "gcc":                      string,
      "build":                    string,
      "flags":                    string/null,
      "date":                     string
    }

Examples:

    $ cat /proc/version | jc --proc -p
    {
      "version": "5.8.0-63-generic",
      "email": "buildd@lcy01-amd64-028",
      "gcc": "gcc (Ubuntu 10.3.0-1ubuntu1~20.10) 10.3.0, GNU ld (GNU Binutils for Ubuntu) 2.35.1",
      "build": "#71-Ubuntu",
      "flags": "SMP",
      "date": "Tue Jul 13 15:59:12 UTC 2021"
    }
"""
import re
from typing import Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/version` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


# inspired by https://gist.github.com/ty2/ad61340e7a4067def2e3c709496bca9d
version_pattern = re.compile(r'''
    Linux\ version\ (?P<version>\S+)\s
    \((?P<email>\S+?)\)\s
    \((?P<gcc>gcc.+)\)\s
    (?P<build>\#\d+(\S+)?)\s
    (?P<flags>.*)?
    (?P<date>(Sun|Mon|Tue|Wed|Thu|Fri|Sat).+)
    ''', re.VERBOSE
)


def _process(proc_data: Dict) -> Dict:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary. Structured to conform to the schema.
    """
    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> Dict:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: Dict = {}

    if jc.utils.has_data(data):
        version_match = version_pattern.match(data)

        if version_match:

            ver_dict = version_match.groupdict()
            raw_output = {x: y.strip() or None for x, y in ver_dict.items()}

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert `du` command output parser

The `du -h` option is not supported with the default output. If you
would like to use `du -h` or other options that change the output, be sure
to use `jc --raw` (cli) or `raw=True` (module).

Usage (cli):

    $ du | jc --du

or

    $ jc du

Usage (module):

    import jc
    result = jc.parse('du', du_command_output)

Schema:

    [
      {
        "size":     integer,
        "name":     string
      }
    ]

Examples:

    $ du /usr | jc --du -p
    [
      {
        "size": 104608,
        "name": "/usr/bin"
      },
      {
        "size": 56,
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      {
        "size": 0,
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      {
        "size": 0,
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      {
        "size": 0,
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      {
        "size": 1008,
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      ...
    ]

    $ du /usr | jc --du -p -r
    [
      {
        "size": "104608",
        "name": "/usr/bin"
      },
      {
        "size": "56",
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      {
        "size": "0",
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      {
        "size": "0",
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      {
        "size": "0",
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      {
        "size": "1008",
        "name": "/usr/standalone/firmware/iBridge1_1Customer.bundle/..."
      },
      ...
    ]
"""
import jc.utils
import jc.parsers.universal


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.6'
    description = '`du` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'aix', 'freebsd']
    magic_commands = ['du']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """
    int_list = {'size'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = []

    # Clear any blank lines
    cleandata = list(filter(None, data.splitlines()))

    if jc.utils.has_data(data):

        cleandata.insert(0, 'size name')
        raw_output = jc.parsers.universal.simple_table_parse(cleandata)

    if raw:
        return raw_output
    else:
        return _process(raw_output)



"""jc - JSON Convert `/proc/buddyinfo` file parser

Usage (cli):

    $ cat /proc/buddyinfo | jc --proc

or

    $ jc /proc/buddyinfo

or

    $ cat /proc/buddyinfo | jc --proc-buddyinfo

Usage (module):

    import jc
    result = jc.parse('proc', proc_buddyinfo_file)

or

    import jc
    result = jc.parse('proc_buddyinfo', proc_buddyinfo_file)

Schema:

All values are integers.

    [
      {
        "node":               integer,
        "zone":               string,
        "free_chunks": [
                              integer  # [0]
        ]
      }
    ]

    [0] array index correlates to the Order number.
        E.g. free_chunks[0] is the value for Order 0


Examples:

    $ cat /proc/buddyinfo | jc --proc -p
    [
      {
        "node": 0,
        "zone": "DMA",
        "free_chunks": [
          0,
          0,
          0,
          1,
          1,
          1,
          1,
          1,
          0,
          1,
          3
        ]
      },
      {
        "node": 0,
        "zone": "DMA32",
        "free_chunks": [
          78,
          114,
          82,
          52,
          38,
          25,
          13,
          9,
          3,
          4,
          629
        ]
      },
      {
        "node": 0,
        "zone": "Normal",
        "free_chunks": [
          0,
          22,
          8,
          10,
          1,
          1,
          2,
          11,
          13,
          0,
          0
        ]
      }
    ]
"""
from typing import List, Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/buddyinfo` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'node'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

        if 'free_chunks' in entry:
            entry['free_chunks'] = [int(x) for x in entry['free_chunks']]

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):

            buddy_list = line.split()

            raw_output.append(
                {
                    'node': buddy_list[1][:-1],
                    'zone': buddy_list[3],
                    'free_chunks': buddy_list[4:]
                }
            )

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert `/proc/net/dev_mcast` file parser

Usage (cli):

    $ cat /proc/net/dev_mcast | jc --proc

or

    $ jc /proc/net/dev_mcast

or

    $ cat /proc/net/dev_mcast | jc --proc-net-dev-mcast

Usage (module):

    import jc
    result = jc.parse('proc', proc_net_dev_mcast_file)

or

    import jc
    result = jc.parse('proc_net_dev_mcast', proc_net_dev_mcast_file)

Schema:

    [
      {
        "index":                      integer,
        "interface":                  string,
        "dmi_u":                      integer,
        "dmi_g":                      integer,
        "dmi_address":                string
      }
    ]

Examples:

    $ cat /proc/net/dev_mcast | jc --proc -p
    [
      {
        "index": 2,
        "interface": "ens33",
        "dmi_u": 1,
        "dmi_g": 0,
        "dmi_address": "333300000001"
      },
      {
        "index": 2,
        "interface": "ens33",
        "dmi_u": 1,
        "dmi_g": 0,
        "dmi_address": "01005e000001"
      },
      ...
    ]

    $ cat /proc/net/dev_mcast | jc --proc-net-dev-mcast -p -r
    [
      {
        "index": "2",
        "interface": "ens33",
        "dmi_u": "1",
        "dmi_g": "0",
        "dmi_address": "333300000001"
      },
      {
        "index": "2",
        "interface": "ens33",
        "dmi_u": "1",
        "dmi_g": "0",
        "dmi_address": "01005e000001"
      },
      ...
    ]
"""
from typing import List, Dict
import jc.utils
from jc.parsers.universal import simple_table_parse


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/net/dev_mcast` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    no_convert = {'interface_name', 'dmi_address'}

    for item in proc_data:
        for key, val in item.items():
            if key not in no_convert:
                try:
                    item[key] = int(val)
                except Exception:
                    pass

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []

    if jc.utils.has_data(data):

        header = 'index interface dmi_u dmi_g dmi_address\n'
        data = header + data
        data_splitlines = data.splitlines()
        raw_output = simple_table_parse(data_splitlines)

    return raw_output if raw else _process(raw_output)



"""jc - JSON Convert `/etc/shadow` file parser

Usage (cli):

    $ sudo cat /etc/shadow | jc --shadow

Usage (module):

    import jc
    result = jc.parse('shadow', shadow_file_output)

Schema:

    [
      {
        "username":       string,
        "password":       string,
        "last_changed":   integer,
        "minimum":        integer,
        "maximum":        integer,
        "warn":           integer,
        "inactive":       integer,
        "expire":         integer
      }
    ]

Examples:

    $ sudo cat /etc/shadow | jc --shadow -p
    [
      {
        "username": "root",
        "password": "*",
        "last_changed": 18113,
        "minimum": 0,
        "maximum": 99999,
        "warn": 7,
        "inactive": null,
        "expire": null
      },
      {
        "username": "daemon",
        "password": "*",
        "last_changed": 18113,
        "minimum": 0,
        "maximum": 99999,
        "warn": 7,
        "inactive": null,
        "expire": null
      },
      {
        "username": "bin",
        "password": "*",
        "last_changed": 18113,
        "minimum": 0,
        "maximum": 99999,
        "warn": 7,
        "inactive": null,
        "expire": null
      },
      ...
    ]

    $ sudo cat /etc/shadow | jc --shadow -p -r
    [
      {
        "username": "root",
        "password": "*",
        "last_changed": "18113",
        "minimum": "0",
        "maximum": "99999",
        "warn": "7",
        "inactive": "",
        "expire": ""
      },
      {
        "username": "daemon",
        "password": "*",
        "last_changed": "18113",
        "minimum": "0",
        "maximum": "99999",
        "warn": "7",
        "inactive": "",
        "expire": ""
      },
      {
        "username": "bin",
        "password": "*",
        "last_changed": "18113",
        "minimum": "0",
        "maximum": "99999",
        "warn": "7",
        "inactive": "",
        "expire": ""
      },
      ...
    ]
"""
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.5'
    description = '`/etc/shadow` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'aix', 'freebsd']
    tags = ['file']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """
    int_list = {'last_changed', 'minimum', 'maximum', 'warn', 'inactive', 'expire'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = []

    # Clear any blank lines
    cleandata = list(filter(None, data.splitlines()))

    if jc.utils.has_data(data):

        for entry in cleandata:
            if entry.startswith('#'):
                continue

            output_line = {}
            fields = entry.split(':')

            output_line['username'] = fields[0]
            output_line['password'] = fields[1]
            output_line['last_changed'] = fields[2]
            output_line['minimum'] = fields[3]
            output_line['maximum'] = fields[4]
            output_line['warn'] = fields[5]
            output_line['inactive'] = fields[6]
            output_line['expire'] = fields[7]

            raw_output.append(output_line)

    if raw:
        return raw_output
    else:
        return _process(raw_output)



"""jc - JSON Convert `uname -a` command output parser

> Note: Must use `uname -a`

Usage (cli):

    $ uname -a | jc --uname

or

    $ jc uname -a

Usage (module):

    import jc
    result = jc.parse('uname', uname_command_output)

Schema:

    {
        "kernel_name":        string,
        "node_name":          string,
        "kernel_release":     string,
        "operating_system":   string,
        "hardware_platform":  string,
        "processor":          string,
        "machine":            string,
        "kernel_version":     string
    }

Example:

    $ uname -a | jc --uname -p
    {
      "kernel_name": "Linux",
      "node_name": "user-ubuntu",
      "kernel_release": "4.15.0-65-generic",
      "operating_system": "GNU/Linux",
      "hardware_platform": "x86_64",
      "processor": "x86_64",
      "machine": "x86_64",
      "kernel_version": "#74-Ubuntu SMP Tue Sep 17 17:06:04 UTC 2019"
    }
"""
import jc.utils
from jc.exceptions import ParseError


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.7'
    description = '`uname -a` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'freebsd']
    magic_commands = ['uname']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary. Structured data to conform to the schema.
    """
    # nothing to process
    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = {}

    if jc.utils.has_data(data):

        # check for macOS or FreeBSD output
        if data.startswith('Darwin') or data.startswith('FreeBSD'):
            parsed_line = data.split()

            if len(parsed_line) < 5:
                raise ParseError('Could not parse uname output. Make sure to use "uname -a".')

            raw_output['machine'] = parsed_line.pop(-1)
            raw_output['kernel_name'] = parsed_line.pop(0)
            raw_output['node_name'] = parsed_line.pop(0)
            raw_output['kernel_release'] = parsed_line.pop(0)
            raw_output['kernel_version'] = ' '.join(parsed_line)

        # otherwise use linux parser
        else:
            # fixup for cases where 'machine' exists but 'processor' and 'hardware_platform' fields are blank.
            # if the fields exist then at least two of the three will be the same.
            # if the fields do not exist then none of the fields in those positions will be the same.
            # case of only two existing is undefined. Must either be one or all three existing, otherwise
            # there will be unexpected results during parsing.
            fixup = data.split()
            if len(fixup) >= 4:
                fixup_set = set([fixup[-2], fixup[-3], fixup[-4]])
                if len(fixup_set) > 2:
                    fixup.insert(-1, 'unknown')
                    fixup.insert(-1, 'unknown')
                    data = ' '.join(fixup)

            parsed_line = data.split(maxsplit=3)

            if len(parsed_line) < 3:
                raise ParseError('Could not parse uname output. Make sure to use "uname -a".')

            raw_output['kernel_name'] = parsed_line.pop(0)
            raw_output['node_name'] = parsed_line.pop(0)
            raw_output['kernel_release'] = parsed_line.pop(0)

            parsed_line = parsed_line[-1].rsplit(maxsplit=4)

            raw_output['operating_system'] = parsed_line.pop(-1)
            raw_output['processor'] = parsed_line.pop(-1)
            raw_output['hardware_platform'] = parsed_line.pop(-1)
            raw_output['machine'] = parsed_line.pop(-1)

            raw_output['kernel_version'] = parsed_line.pop(0)

    if raw:
        return raw_output
    else:
        return _process(raw_output)



"""jc - JSON Convert `dmidecode` command output parser

Usage (cli):

    $ dmidecode | jc --dmidecode

or

    $ jc dmidecode

Usage (module):

    import jc
    result = jc.parse('dmidecode', dmidecode_command_output)

Schema:

    [
      {
        "handle":                      string,
        "type":                        integer,
        "bytes":                       integer,
        "description":                 string,
        "values": {                                 # null if empty
          "lowercase_no_spaces_keys":  string,
          "multiline_key_values": [
                                       string,
          ]
        }
      }
    ]

Examples:

    # dmidecode | jc --dmidecode -p
    [
      {
        "handle": "0x0000",
        "type": 0,
        "bytes": 24,
        "description": "BIOS Information",
        "values": {
          "vendor": "Phoenix Technologies LTD",
          "version": "6.00",
          "release_date": "04/13/2018",
          "address": "0xEA490",
          "runtime_size": "88944 bytes",
          "rom_size": "64 kB",
          "characteristics": [
            "ISA is supported",
            "PCI is supported",
            "PC Card (PCMCIA) is supported",
            "PNP is supported",
            "APM is supported",
            "BIOS is upgradeable",
            "BIOS shadowing is allowed",
            "ESCD support is available",
            "Boot from CD is supported",
            "Selectable boot is supported",
            "EDD is supported",
            "Print screen service is supported (int 5h)",
            "8042 keyboard services are supported (int 9h)",
            "Serial services are supported (int 14h)",
            "Printer services are supported (int 17h)",
            "CGA/mono video services are supported (int 10h)",
            "ACPI is supported",
            "Smart battery is supported",
            "BIOS boot specification is supported",
            "Function key-initiated network boot is supported",
            "Targeted content distribution is supported"
          ],
          "bios_revision": "4.6",
          "firmware_revision": "0.0"
        }
      },
      ...
    ]

    # dmidecode | jc --dmidecode -p -r
    [
      {
        "handle": "0x0000",
        "type": "0",
        "bytes": "24",
        "description": "BIOS Information",
        "values": {
          "vendor": "Phoenix Technologies LTD",
          "version": "6.00",
          "release_date": "04/13/2018",
          "address": "0xEA490",
          "runtime_size": "88944 bytes",
          "rom_size": "64 kB",
          "characteristics": [
            "ISA is supported",
            "PCI is supported",
            "PC Card (PCMCIA) is supported",
            "PNP is supported",
            "APM is supported",
            "BIOS is upgradeable",
            "BIOS shadowing is allowed",
            "ESCD support is available",
            "Boot from CD is supported",
            "Selectable boot is supported",
            "EDD is supported",
            "Print screen service is supported (int 5h)",
            "8042 keyboard services are supported (int 9h)",
            "Serial services are supported (int 14h)",
            "Printer services are supported (int 17h)",
            "CGA/mono video services are supported (int 10h)",
            "ACPI is supported",
            "Smart battery is supported",
            "BIOS boot specification is supported",
            "Function key-initiated network boot is supported",
            "Targeted content distribution is supported"
          ],
          "bios_revision": "4.6",
          "firmware_revision": "0.0"
        }
      },
      ...
    ]
"""
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.5'
    description = '`dmidecode` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    magic_commands = ['dmidecode']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """
    int_list = {'type', 'bytes'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

        if not entry['values']:
            entry['values'] = None

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    item_header = False
    item_values = False
    value_list = False

    item = None
    header = None
    key = None
    val = None
    attribute = None
    values = None
    key_data = None

    raw_output = []

    if jc.utils.has_data(data):

        data = data.splitlines()

        # remove header rows
        for row in data.copy():
            if row:
                data.pop(0)
            else:
                break

        # main parsing loop
        for line in data:
            # new item
            if not line:
                item_header = True
                item_values = False
                value_list = False

                if item:
                    if values:
                        item['values'][attribute] = values
                    if key_data:
                        item['values'][f'{key}_data'] = key_data
                    raw_output.append(item)

                item = {}
                header = None
                key = None
                val = None
                attribute = None
                values = []
                key_data = []
                continue

            # header
            if line.startswith('Handle ') and line.endswith('bytes'):

                # Handle 0x0000, DMI type 0, 24 bytes
                header = line.replace(',', ' ').split()
                item = {
                    'handle': header[1],
                    'type': header[4],
                    'bytes': header[5]
                }
                continue

            # description
            if item_header:
                item_header = False
                item_values = True
                value_list = False

                item['description'] = line
                item['values'] = {}
                continue

            # new item if multiple descriptions in handle
            if not item_header and not line.startswith('\t'):
                item_header = False
                item_values = True
                value_list = False

                if item:
                    if values:
                        item['values'][attribute] = values
                    if key_data:
                        item['values'][f'{key}_data'] = key_data
                    raw_output.append(item)

                item = {
                    'handle': header[1],
                    'type': header[4],
                    'bytes': header[5],
                    'description': line,
                    'values': {}
                }

                key = None
                val = None
                attribute = None
                values = []
                key_data = []
                continue

            # keys and values
            if item_values \
               and len(line.split(':', maxsplit=1)) == 2 \
               and line.startswith('\t') \
               and not line.startswith('\t\t') \
               and not line.strip().endswith(':'):
                item_header = False
                item_values = True
                value_list = False

                if values:
                    item['values'][attribute] = values
                    values = []
                if key_data:
                    item['values'][f'{key}_data'] = key_data
                    key_data = []

                key = line.split(':', maxsplit=1)[0].strip().lower().replace(' ', '_')
                val = line.split(':', maxsplit=1)[1].strip()
                item['values'].update({key: val})
                continue

            # multi-line key
            if item_values \
               and line.startswith('\t') \
               and not line.startswith('\t\t') \
               and line.strip().endswith(':'):
                item_header = False
                item_values = True
                value_list = True

                if values:
                    item['values'][attribute] = values
                    values = []
                if key_data:
                    item['values'][f'{key}_data'] = key_data
                    key_data = []

                attribute = line[:-1].strip().lower().replace(' ', '_')
                values = []
                continue

            # multi-line values
            if value_list \
               and line.startswith('\t\t'):
                values.append(line.strip())
                continue

            # data for hybrid multi-line objects
            if item_values \
               and not value_list \
               and line.startswith('\t\t'):
                if f'{key}_data' not in item['values']:
                    item['values'][f'{key}_data'] = []
                key_data.append(line.strip())
                continue

        if item:
            raw_output.append(item)

    if raw:
        return raw_output
    else:
        return _process(raw_output)



"""jc - JSON Convert utils"""
import sys
import re
import locale
import shutil
from datetime import datetime, timezone
from textwrap import TextWrapper
from functools import lru_cache
from typing import Any, List, Dict, Iterable, Union, Optional, TextIO
from .jc_types import TimeStampFormatType


def _asciify(string: str) -> str:
    """
    Return a string downgraded from Unicode to ASCII with some simple
    conversions.
    """
    string = string.replace('©', '(c)')
    # the ascii() function adds single quotes around the string
    string = ascii(string)[1:-1]
    string = string.replace(r'\n', '\n')
    return string


def _safe_print(
    string: str,
    sep: str = ' ',
    end: str = '\n',
    file: TextIO = sys.stdout,
    flush: bool = False
) -> None:
    """Output for both UTF-8 and ASCII encoding systems"""
    try:
        print(string, sep=sep, end=end, file=file, flush=flush)
    except UnicodeEncodeError:
        print(_asciify(string), sep=sep, end=end, file=file, flush=flush)


def _safe_pager(string: str) -> None:
    """Pager output for both UTF-8 and ASCII encoding systems"""
    from pydoc import pager
    try:
        pager(string)
    except UnicodeEncodeError:
        pager(_asciify(string))


def warning_message(message_lines: List[str]) -> None:
    """
    Prints warning message to `STDERR` for non-fatal issues. The first line
    is prepended with 'jc:  Warning - ' and subsequent lines are indented.
    Wraps text as needed based on the terminal width.

    Parameters:

        message:   (list) list of string lines

    Returns:

        None - just prints output to STDERR
    """
    # this is for backwards compatibility with existing custom parsers
    if isinstance(message_lines, str):
        message_lines = [message_lines]

    columns = shutil.get_terminal_size().columns

    first_wrapper = TextWrapper(width=columns, subsequent_indent=' ' * 15)
    next_wrapper = TextWrapper(width=columns, initial_indent=' ' * 15,
                               subsequent_indent=' ' * 19)

    first_line = message_lines.pop(0)
    first_str = f'jc:  Warning - {first_line}'
    first_str = first_wrapper.fill(first_str)
    _safe_print(first_str, file=sys.stderr)

    for line in message_lines:
        if line == '':
            continue
        message = next_wrapper.fill(line)
        _safe_print(message, file=sys.stderr)


def error_message(message_lines: List[str]) -> None:
    """
    Prints an error message to `STDERR` for fatal issues. The first line is
    prepended with 'jc:  Error - ' and subsequent lines are indented.
    Wraps text as needed based on the terminal width.

    Parameters:

        message:   (list) list of string lines

    Returns:

        None - just prints output to STDERR
    """
    columns = shutil.get_terminal_size().columns

    first_wrapper = TextWrapper(width=columns, subsequent_indent=' ' * 13)
    next_wrapper = TextWrapper(width=columns, initial_indent=' ' * 13,
                               subsequent_indent=' ' * 17)

    first_line = message_lines.pop(0)
    first_str = f'jc:  Error - {first_line}'
    first_str = first_wrapper.fill(first_str)
    _safe_print(first_str, file=sys.stderr)

    for line in message_lines:
        if line == '':
            continue
        message = next_wrapper.fill(line)
        _safe_print(message, file=sys.stderr)


def is_compatible(compatible: List[str]) -> bool:
    """
    Returns True if the parser is compatible with the running OS platform.
    """
    platform_found = False

    for platform in compatible:
        if sys.platform.startswith(platform):
            platform_found = True
            break

    return platform_found


def compatibility(mod_name: str, compatible: List[str], quiet: bool = False) -> None:
    """
    Checks for the parser's compatibility with the running OS platform and
    prints a warning message to `STDERR` if not compatible and
    `quiet=False.`

    Parameters:

        mod_name:     (string) __name__ of the calling module

        compatible:   (list) sys.platform name(s) compatible with
                      the parser. compatible options:
                      linux, darwin, cygwin, win32, aix, freebsd

        quiet:        (bool) suppress compatibility message if True

    Returns:

        None - just prints output to STDERR
    """
    if not quiet and not is_compatible(compatible):
        mod = mod_name.split('.')[-1]
        compat_list = ', '.join(compatible)
        warning_message([
            f'`{mod}` command output from this OS ({sys.platform}) is not supported.',
            f'`{mod}` command output from the following platforms is supported: {compat_list}',
            'Disregard this warning if you are processing output that came from a supported platform. (Use the -q option to suppress this warning)'
        ])


def has_data(data: Union[str, bytes]) -> bool:
    """
    Checks if the string input contains data. If there are any
    non-whitespace characters then return `True`, else return `False`.

    For bytes, returns True if there is any data.

    Parameters:

        data:        (string, bytes) input to check whether it contains data

    Returns:

        Boolean      True if input string (data) contains non-whitespace
                     characters, otherwise False. For bytes data, returns
                     True if there is any data, otherwise False.
    """
    if isinstance(data, str):
        return bool(data and not data.isspace())

    return bool(data)


def convert_to_int(value: object) -> Optional[int]:
    """
    Converts string and float input to int. Strips all non-numeric
    characters from strings.

    Parameters:

        value:         (string/float) Input value

    Returns:

        integer/None   Integer if successful conversion, otherwise None
    """
    if isinstance(value, str):
        str_val = re.sub(r'[^0-9\-\.]', '', value)
        try:
            return int(str_val)
        except (ValueError, TypeError):
            try:
                return int(float(str_val))
            except (ValueError, TypeError):
                return None

    elif isinstance(value, (int, float)):
        return int(value)

    else:
        return None


def convert_to_float(value: object) -> Optional[float]:
    """
    Converts string and int input to float. Strips all non-numeric
    characters from strings.

    Parameters:

        value:         (string/integer) Input value

    Returns:

        float/None     Float if successful conversion, otherwise None
    """
    if isinstance(value, str):
        try:
            return float(re.sub(r'[^0-9\-\.]', '', value))
        except (ValueError, TypeError):
            return None

    elif isinstance(value, (int, float)):
        return float(value)

    else:
        return None


def convert_to_bool(value: object) -> bool:
    """
    Converts string, integer, or float input to boolean by checking
    for 'truthy' values.

    Parameters:

        value:          (string/integer/float) Input value

    Returns:

        True/False      False unless a 'truthy' number or string is found
                        ('y', 'yes', 'true', '1', 1, -1, etc.)
    """
    # if number, then bool it
    # if string, try to convert to float
    #   if float converts, then bool the result
    #   if float does not convert then look for truthy string and bool True
    #   else False
    truthy = ['y', 'yes', 'true', '*']

    if isinstance(value, (int, float)):
        return bool(value)

    if isinstance(value, str):
        try:
            test_value = convert_to_float(value)
            if test_value is not None:
                return bool(test_value)
        except Exception:
            pass

        if value:
            return value.lower() in truthy

    return False


def input_type_check(data: object) -> None:
    """Ensure input data is a string. Raises `TypeError` if not."""
    if not isinstance(data, str):
        raise TypeError("Input data must be a 'str' object.")


class timestamp:
    __slots__ = ('string', 'format', 'naive', 'utc', 'iso')

    def __init__(self,
                 datetime_string: Optional[str],
                 format_hint: Optional[Iterable[int]] = None
    ) -> None:
        """
        Input a datetime text string of several formats and convert to a
        naive or timezone-aware epoch timestamp in UTC.

        Parameters:

            datetime_string  (str):  a string representation of a
                datetime in several supported formats

            format_hint  (iterable):  an optional iterable of format ID
                integers to instruct the timestamp object to try those
                formats first in the order given. Other formats will be
                tried after the format hint list is exhausted. This can
                speed up timestamp conversion so several different formats
                don't have to be tried in brute-force fashion.

        Returns a timestamp object with the following attributes:

            string  (str):  the input datetime string

            format  (int | None):  the format rule that was used to decode
                the datetime string. None if conversion fails.

            naive  (int | None):  timestamp based on locally configured
                timezone. None if conversion fails.

            utc  (int | None):  aware timestamp only if UTC timezone
                detected in datetime string. None if conversion fails.

            iso (str | None):  ISO string - timezone information is output
                only if UTC timezone is detected in the datetime string.
        """
        self.string = datetime_string

        if not format_hint:
            format_hint = tuple()
        else:
            format_hint = tuple(format_hint)

        dt = self._parse_dt(self.string, format_hint=format_hint)
        self.format = dt['format']
        self.naive = dt['timestamp_naive']
        self.utc = dt['timestamp_utc']
        self.iso = dt['iso']

    def __repr__(self) -> str:
        return f'timestamp(string={self.string!r}, format={self.format}, naive={self.naive}, utc={self.utc}, iso={self.iso!r})'

    @staticmethod
    @lru_cache(maxsize=2048)
    def _parse_dt(
        dt_string: Optional[str],
        format_hint: Optional[Iterable[int]] = None
    ) -> Dict[str, Any]:
        """
        Input a datetime text string of several formats and convert to
        a naive or timezone-aware epoch timestamp in UTC.

        Parameters:

            dt_string:    (string) a string representation of a date-time
                          in several supported formats

            format_hint:  (list | tuple) a list of format ID int's that
                          should be tried first. This can increase
                          performance since the function will not need to
                          try many incorrect formats before finding the
                          correct one.

        Returns:

            Dictionary of the following format:

                {
                    # for debugging purposes. None if conversion fails
                    "format":               int,

                    # timestamp based on locally configured timezone.
                    # None if conversion fails.
                    "timestamp_naive":      int,

                    # aware timestamp only if UTC timezone detected.
                    # None if conversion fails.
                    "timestamp_utc":        int

                    # ISO string. None if conversion fails.
                    "iso":                  str
                }

                The `format` integer denotes which date_time format
                conversion succeeded.

                The `timestamp_naive` integer is the converted date-time
                string to a naive epoch timestamp.

                The `timestamp_utc` integer is the converted date-time
                string to an aware epoch timestamp in the UTC timezone. If
                an aware conversion cannot be performed (e.g. the UTC
                timezone is not found in the date-time string), then this
                field will be None.

                The `iso` string will only have timezone information if the
                UTC timezone is detected in `dt_string`.

                If the conversion completely fails, all fields will be None.
        """
        formats: tuple[TimeStampFormatType, ...] = (
            {'id': 1000, 'format': '%a %b %d %H:%M:%S %Y', 'locale': None},  # manual C locale format conversion: Tue Mar 23 16:12:11 2021 or Tue Mar 23 16:12:11 IST 2021
            {'id': 1100, 'format': '%a %b %d %H:%M:%S %Y %z', 'locale': None}, # git date output: Thu Mar 5 09:17:40 2020 -0800
            {'id': 1300, 'format': '%Y-%m-%dT%H:%M:%S.%f%Z', 'locale': None}, # ISO Format with UTC (found in syslog 5424): 2003-10-11T22:14:15.003Z
            {'id': 1310, 'format': '%Y-%m-%dT%H:%M:%S.%f', 'locale': None}, # ISO Format without TZ (found in syslog 5424): 2003-10-11T22:14:15.003
            {'id': 1400, 'format': '%b %d %Y %H:%M:%S.%f UTC', 'locale': None}, # CEF Format with UTC: Nov 08 2022 12:30:00.111 UTC
            {'id': 1410, 'format': '%b %d %Y %H:%M:%S.%f', 'locale': None}, # CEF Format without TZ: Nov 08 2022 12:30:00.111
            {'id': 1420, 'format': '%b %d %Y %H:%M:%S UTC', 'locale': None}, # CEF Format with UTC without microseconds: Nov 08 2022 12:30:00 UTC
            {'id': 1430, 'format': '%b %d %Y %H:%M:%S', 'locale': None}, # CEF Format without TZ or microseconds: Nov 08 2022 12:30:00
            {'id': 1500, 'format': '%Y-%m-%d %H:%M', 'locale': None},  # en_US.UTF-8 local format (found in who cli output): 2021-03-23 00:14
            {'id': 1600, 'format': '%m/%d/%Y %I:%M %p', 'locale': None},  # Windows english format (found in dir cli output): 12/07/2019 02:09 AM
            {'id': 1700, 'format': '%m/%d/%Y, %I:%M:%S %p', 'locale': None},  # Windows english format wint non-UTC tz (found in systeminfo cli output): 3/22/2021, 1:15:51 PM (UTC-0600)
            {'id': 1705, 'format': '%m/%d/%Y, %I:%M:%S %p %Z', 'locale': None},  # Windows english format with UTC tz (found in systeminfo cli output): 3/22/2021, 1:15:51 PM (UTC)
            {'id': 1710, 'format': '%m/%d/%Y, %I:%M:%S %p UTC%z', 'locale': None},  # Windows english format with UTC tz (found in systeminfo cli output): 3/22/2021, 1:15:51 PM (UTC+0000)
            {'id': 1750, 'format': '%Y/%m/%d-%H:%M:%S.%f', 'locale': None},  # Google Big Table format with no timezone: 1970/01/01-01:00:00.000000
            {'id': 1755, 'format': '%Y/%m/%d-%H:%M:%S.%f%z', 'locale': None},  # Google Big Table format with timezone: 1970/01/01-01:00:00.000000+00:00
            {'id': 1760, 'format': '%Y-%m-%d %H:%M:%S%z', 'locale': None},  # certbot format with timezone: 2023-06-12 01:35:30+00:00
            {'id': 1800, 'format': '%d/%b/%Y:%H:%M:%S %z', 'locale': None},  # Common Log Format: 10/Oct/2000:13:55:36 -0700
            {'id': 2000, 'format': '%a %d %b %Y %I:%M:%S %p %Z', 'locale': None},  # en_US.UTF-8 local format (found in upower cli output): Tue 23 Mar 2021 04:12:11 PM UTC
            {'id': 3000, 'format': '%a %d %b %Y %I:%M:%S %p', 'locale': None},  # en_US.UTF-8 local format with non-UTC tz (found in upower cli output): Tue 23 Mar 2021 04:12:11 PM IST
            {'id': 4000, 'format': '%A %d %B %Y %I:%M:%S %p %Z', 'locale': None},  # European-style local format (found in upower cli output): Tuesday 01 October 2019 12:50:41 PM UTC
            {'id': 5000, 'format': '%A %d %B %Y %I:%M:%S %p', 'locale': None},  # European-style local format with non-UTC tz (found in upower cli output): Tuesday 01 October 2019 12:50:41 PM IST
            {'id': 6000, 'format': '%a %b %d %I:%M:%S %p %Z %Y', 'locale': None},  # en_US.UTF-8 format (found in date cli): Wed Mar 24 06:16:19 PM UTC 2021
            {'id': 7000, 'format': '%a %b %d %H:%M:%S %Z %Y', 'locale': None},  # C locale format (found in date cli): Wed Mar 24 11:11:30 UTC 2021
            {'id': 7100, 'format': '%b %d %H:%M:%S %Y', 'locale': None},  # C locale format (found in stat cli output - osx): # Mar 29 11:49:05 2021
            {'id': 7200, 'format': '%Y-%m-%d %H:%M:%S.%f %z', 'locale': None},  # C locale format (found in stat cli output - linux): 2019-08-13 18:13:43.555604315 -0400
            {'id': 7250, 'format': '%Y-%m-%d %H:%M:%S', 'locale': None},  # C locale format with non-UTC tz (found in modified vmstat cli output): # 2021-09-16 20:32:28 PDT
            {'id': 7255, 'format': '%Y-%m-%d %H:%M:%S %Z', 'locale': None},  # C locale format (found in modified vmstat cli output): # 2021-09-16 20:32:28 UTC
            {'id': 7300, 'format': '%a %Y-%m-%d %H:%M:%S %Z', 'locale': None},  # C locale format (found in timedatectl cli output): # Wed 2020-03-11 00:53:21 UTC
            # attempt locale changes last
            {'id': 8000, 'format': '%a %d %b %Y %H:%M:%S %Z', 'locale': ''},  # current locale format (found in upower cli output): # mar. 23 mars 2021 23:12:11 UTC
            {'id': 8100, 'format': '%a %d %b %Y %H:%M:%S', 'locale': ''},  # current locale format with non-UTC tz (found in upower cli output): # mar. 23 mars 2021 19:12:11 EDT
            {'id': 8200, 'format': '%A %d %B %Y, %H:%M:%S UTC%z', 'locale': ''},  # fr_FR.utf8 locale format (found in date cli output): vendredi 26 mars 2021, 13:26:46 (UTC+0000)
            {'id': 8300, 'format': '%A %d %B %Y, %H:%M:%S', 'locale': ''},  # fr_FR.utf8 locale format with non-UTC tz (found in date cli output): vendredi 26 mars 2021, 13:26:46 (UTC-0400)
            {'id': 9000, 'format': '%c', 'locale': ''}  # locally configured locale format conversion: Could be anything :) this is a last-gasp attempt
        )

        # from https://www.timeanddate.com/time/zones/
        # only removed UTC timezone and added known non-UTC offsets
        tz_abbr: set[str] = {
            'A', 'ACDT', 'ACST', 'ACT', 'ACWST', 'ADT', 'AEDT', 'AEST', 'AET', 'AFT', 'AKDT',
            'AKST', 'ALMT', 'AMST', 'AMT', 'ANAST', 'ANAT', 'AQTT', 'ART', 'AST', 'AT', 'AWDT',
            'AWST', 'AZOST', 'AZOT', 'AZST', 'AZT', 'AoE', 'B', 'BNT', 'BOT', 'BRST', 'BRT', 'BST',
            'BTT', 'C', 'CAST', 'CAT', 'CCT', 'CDT', 'CEST', 'CET', 'CHADT', 'CHAST', 'CHOST',
            'CHOT', 'CHUT', 'CIDST', 'CIST', 'CKT', 'CLST', 'CLT', 'COT', 'CST', 'CT', 'CVT', 'CXT',
            'ChST', 'D', 'DAVT', 'DDUT', 'E', 'EASST', 'EAST', 'EAT', 'ECT', 'EDT', 'EEST', 'EET',
            'EGST', 'EGT', 'EST', 'ET', 'F', 'FET', 'FJST', 'FJT', 'FKST', 'FKT', 'FNT', 'G',
            'GALT', 'GAMT', 'GET', 'GFT', 'GILT', 'GMT', 'GST', 'GYT', 'H', 'HDT', 'HKT', 'HOVST',
            'HOVT', 'HST', 'I', 'ICT', 'IDT', 'IOT', 'IRDT', 'IRKST', 'IRKT', 'IRST', 'IST', 'JST',
            'K', 'KGT', 'KOST', 'KRAST', 'KRAT', 'KST', 'KUYT', 'L', 'LHDT', 'LHST', 'LINT', 'M',
            'MAGST', 'MAGT', 'MART', 'MAWT', 'MDT', 'MHT', 'MMT', 'MSD', 'MSK', 'MST', 'MT', 'MUT',
            'MVT', 'MYT', 'N', 'NCT', 'NDT', 'NFDT', 'NFT', 'NOVST', 'NOVT', 'NPT', 'NRT', 'NST',
            'NUT', 'NZDT', 'NZST', 'O', 'OMSST', 'OMST', 'ORAT', 'P', 'PDT', 'PET', 'PETST', 'PETT',
            'PGT', 'PHOT', 'PHT', 'PKT', 'PMDT', 'PMST', 'PONT', 'PST', 'PT', 'PWT', 'PYST', 'PYT',
            'Q', 'QYZT', 'R', 'RET', 'ROTT', 'S', 'SAKT', 'SAMT', 'SAST', 'SBT', 'SCT', 'SGT',
            'SRET', 'SRT', 'SST', 'SYOT', 'T', 'TAHT', 'TFT', 'TJT', 'TKT', 'TLT', 'TMT', 'TOST',
            'TOT', 'TRT', 'TVT', 'U', 'ULAST', 'ULAT', 'UYST', 'UYT', 'UZT', 'V', 'VET', 'VLAST',
            'VLAT', 'VOST', 'VUT', 'W', 'WAKT', 'WARST', 'WAST', 'WAT', 'WEST', 'WET', 'WFT',
            'WGST', 'WGT', 'WIB', 'WIT', 'WITA', 'WST', 'WT', 'X', 'Y', 'YAKST', 'YAKT', 'YAPT',
            'YEKST', 'YEKT', 'UTC-1200', 'UTC-1100', 'UTC-1000', 'UTC-0930', 'UTC-0900',
            'UTC-0800', 'UTC-0700', 'UTC-0600', 'UTC-0500', 'UTC-0400', 'UTC-0300', 'UTC-0230',
            'UTC-0200', 'UTC-0100', 'UTC+0100', 'UTC+0200', 'UTC+0300', 'UTC+0400', 'UTC+0430',
            'UTC+0500', 'UTC+0530', 'UTC+0545', 'UTC+0600', 'UTC+0630', 'UTC+0700', 'UTC+0800',
            'UTC+0845', 'UTC+0900', 'UTC+1000', 'UTC+1030', 'UTC+1100', 'UTC+1200', 'UTC+1300',
            'UTC+1345', 'UTC+1400'
        }

        offset_suffixes: tuple[str, ...] = (
            '-12:00', '-11:00', '-10:00', '-09:30', '-09:00',
            '-08:00', '-07:00', '-06:00', '-05:00', '-04:00', '-03:00', '-02:30',
            '-02:00', '-01:00', '+01:00', '+02:00', '+03:00', '+04:00', '+04:30',
            '+05:00', '+05:30', '+05:45', '+06:00', '+06:30', '+07:00', '+08:00',
            '+08:45', '+09:00', '+10:00', '+10:30', '+11:00', '+12:00', '+13:00',
            '+13:45', '+14:00'
        )

        data: str = dt_string or ''
        normalized_datetime: str = ''
        utc_tz: bool = False
        dt: Optional[datetime] = None
        dt_utc: Optional[datetime] = None
        timestamp_naive: Optional[int] = None
        timestamp_utc: Optional[int] = None
        iso_string: Optional[str] = None
        timestamp_obj: Dict[str, Any] = {
            'format': None,
            'timestamp_naive': None,
            'timestamp_utc': None,
            'iso': None
        }

        # convert format_hint to a tuple so it is hashable (for lru_cache)
        if not format_hint:
            format_hint = tuple()
        else:
            format_hint = tuple(format_hint)

        # sometimes UTC is referenced as 'Coordinated Universal Time'. Convert to 'UTC'
        data = data.replace('Coordinated Universal Time', 'UTC')

        # UTC can also be indicated with 'Z' for Zulu time (ISO-8601). Convert to 'UTC'
        data = data.replace('Z', 'UTC')

        if 'UTC' in data:
            utc_tz = True
            if 'UTC+' in data or 'UTC-' in data:
                utc_tz = bool('UTC+0000' in data or 'UTC-0000' in data)

        elif '+0000' in data \
             or '-0000' in data \
             or '+00:00' in data \
             or '-00:00' in data:
            utc_tz = True
            data = data.replace('+00:00', '+0000')  # fix for python 3.6

        # normalize the timezone by taking out any timezone reference, except UTC
        cleandata = data.replace('(', '').replace(')', '')
        normalized_datetime_list: List[str] = []
        for term in cleandata.split():
            if term not in tz_abbr:
                normalized_datetime_list.append(term)

        normalized_datetime = ' '.join(normalized_datetime_list)

        # remove non UTC offset suffixes at the end of the string
        for suffix in offset_suffixes:
            if normalized_datetime.endswith(suffix):
                normalized_datetime = normalized_datetime[0:-len(suffix)]
                break

        # normalize further by converting any greater-than 6-digit subsecond to 6-digits
        p = re.compile(r'(\W\d\d:\d\d:\d\d\.\d{6})\d+\W')
        normalized_datetime = p.sub(r'\g<1> ', normalized_datetime)

        # try format hints first, then fall back to brute-force method
        hint_obj_list: List[TimeStampFormatType] = []
        for fmt_id in format_hint:
            for fmt in formats:
                if fmt_id == fmt['id']:
                    hint_obj_list.append(fmt)

        remaining_formats = [fmt for fmt in formats if not fmt['id'] in format_hint]
        optimized_formats = hint_obj_list + remaining_formats

        for fmt in optimized_formats:
            try:
                locale.setlocale(locale.LC_TIME, fmt['locale'])
                dt = datetime.strptime(normalized_datetime, fmt['format'])
                timestamp_obj['format'] = fmt['id']
                timestamp_naive = int(dt.replace(tzinfo=None).timestamp())
                iso_string = dt.replace(tzinfo=None).isoformat()
                locale.setlocale(locale.LC_TIME, None)
                break
            except Exception:
                locale.setlocale(locale.LC_TIME, None)
                continue

        if dt and utc_tz:
            dt_utc = dt.replace(tzinfo=timezone.utc)
            timestamp_utc = int(dt_utc.timestamp())
            iso_string = dt_utc.isoformat()

        if timestamp_naive:
            timestamp_obj['timestamp_naive'] = timestamp_naive
            timestamp_obj['timestamp_utc'] = timestamp_utc
            timestamp_obj['iso'] = iso_string

        return timestamp_obj



"""jc - JSON Convert INI with duplicate key file parser

Parses standard INI files and preserves duplicate values. All values are
contained in lists/arrays.

- Delimiter can be `=` or `:`. Missing values are supported.
- Comment prefix can be `#` or `;`. Comments must be on their own line.
- If any section names have the same name as a top-level key, the top-level
  key will be overwritten by the section data.
- If multi-line values are used, each line will be a separate item in the
  value list. Blank lines in multi-line values are not supported.

> Note: Values starting and ending with double or single quotation marks
> will have the marks removed. If you would like to keep the quotation
> marks, use the `-r` command-line argument or the `raw=True` argument in
> `parse()`.

Usage (cli):

    $ cat foo.ini | jc --ini

Usage (module):

    import jc
    result = jc.parse('ini', ini_file_output)

Schema:

INI document converted to a dictionary - see the python configparser
standard library documentation for more details.

    {
      "<key1>": [
                            string
      ],
      "<key2>": [
                            string
      ],
      "<section1>": {
        "<key1>": [
                            string
        ],
        "<key2>": [
                            string
        ]
      }
    }

Examples:

    $ cat example.ini
    foo = fiz
    bar = buz

    [section1]
    fruit = apple
    color = blue
    color = red

    [section2]
    fruit = pear
    fruit = peach
    color = green

    $ cat example.ini | jc --ini -p
    {
      "foo": [
        "fiz"
      ],
      "bar": [
        "buz"
      ],
      "section1": {
        "fruit": [
          "apple"
        ],
        "color": [
          "blue",
          "red"
        ]
      },
      "section2": {
        "fruit": [
          "pear",
          "peach"
        ],
        "color": [
          "green"
        ]
      }
    }
"""
import jc.utils
import configparser
import uuid


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = 'INI with duplicate key file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    details = 'Using configparser from the python standard library'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['standard', 'file', 'string']


__version__ = info.version


class MultiDict(dict):
    # https://stackoverflow.com/a/38286559/12303989
    def __setitem__(self, key, value):
        if key in self:
            if isinstance(value, list):
                self[key].extend(value)

            elif isinstance(value, str):
                if len(self[key])>1:
                    return

        else:
            super().__setitem__(key, value)


def _remove_quotes(value):
    if value is None:
        value = ''

    elif value.startswith('"') and value.endswith('"'):
        value = value[1:-1]

    elif value.startswith("'") and value.endswith("'"):
        value = value[1:-1]

    return value


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary representing the INI file.
    """
    # remove quotation marks from beginning and end of values
    for k, v in proc_data.items():
        if isinstance(v, dict):
            for key, value in v.items():
                if isinstance(value, list):
                    v[key] = [_remove_quotes(x) for x in value]
                else:
                    v[key] = _remove_quotes(value)
            continue

        elif isinstance(v, list):
            proc_data[k] = [_remove_quotes(x) for x in v]

        else:
            proc_data[k] = _remove_quotes(v)

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary representing the INI file.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = {}

    if jc.utils.has_data(data):

        ini_parser = configparser.ConfigParser(
            dict_type = MultiDict,
            allow_no_value=True,
            interpolation=None,
            default_section=None,
            empty_lines_in_values=False,
            strict=False
        )

        # don't convert keys to lower-case:
        ini_parser.optionxform = lambda option: option

        try:
            ini_parser.read_string(data)
            raw_output = {s: dict(ini_parser.items(s)) for s in ini_parser.sections()}

        except configparser.MissingSectionHeaderError:
            # find a top-level section name that will not collide with any existing ones
            while True:
                my_uuid = str(uuid.uuid4())
                if my_uuid not in data:
                    break

            data = f'[{my_uuid}]\n' + data
            ini_parser.read_string(data)
            temp_dict = {s: dict(ini_parser.items(s)) for s in ini_parser.sections()}

            # move items under fake top-level sections to the root
            raw_output = temp_dict.pop(my_uuid)

            # get the rest of the sections
            raw_output.update(temp_dict)

    return raw_output if raw else _process(raw_output)
